;   fltptc65 -- floating point routines for 65C02
;
;   (C) 1999 - 2008, C. Bond. All rights reserved.
;   (C) 2014 - 2017  M. Ferrer for new functions (v.6 below).
;
;   v.1
;   This version includes add, subtract, multiply, divide, square
;   root, tangent and arctangent. The tangent and arctangent are
;   implemented as efficient BCD CORDIC algorithms. Sine, cosine
;   arcsine and arccosine are also provided.
;
;   v.2
;   This version is a massive rewrite of the previous version to
;   simplify and improve the algoritms and structure of the trig
;   functions.
;
;   v.3
;   Added remaining trig and inverse trig functions along with
;   natural log (ln) and exponential (exp).
;
;   v.4
;   Improved error reporting and added hyperbolic and inverse
;   hyperbolic functions.
;
;   v.5
;   Added log2, log10 and power (pow) function.
;
;   Todo:
;       1. Provide better documentation for all code,
;       2. Remove unused code and unreferenced labels.
;
;   v.6
;   * Fixed bugs :
;     mul  : main routine reworked (please see routine's header comment)
;     atan : added null argument processing (caused endless loop)
;     int  : fixed bug when exponent is +010 (erased the 2nd digit
;            of mantissa instead of the 12th)
;   cmpreg : fixed case where signs & exponents equal and
;            mantissas negative
;
;   * Added functions : pow2, isint, cubrt, p10ow
;                       fact, ncr, npr, p2r, r2p, hr2hms, hms2hr
;
;
;
;--------------------------------------------------------------------------
;
;   The internal 8-byte floating point format is as follows:
;
;
;       byte:   0   1   2   3   4   5   6   7
;               Â   Â   Â   Â   Â   Â   Â   Â
;               ³   ³   ³   ³   ³   ³   ³   ³
;               ³   ³   ÀÄÄÄÁÄÄÄÁÄÄÄÁÄÄÄÁÄÄÄÁÄÄÄÄÄ Mantissa (12 BCD DIGITS)
;               ³   ³  D.D  DD  DD  DD  DD  DD
;               ³   ³   ³
;               ³   ³   ÀÄÄÄÄÄÄÄÄÄÄÄÄÄÄÄÄÄÄÄÄÄÄÄÄÄ (implied decimal point)
;               ³   ³ 
;               ³   ÀÄÄÄÄÄÄÄÄÄÄÄÄÄÄÄÄÄÄÄÄÄÄÄÂÄÂÄÄÄ Middle and low 
;               ³                           ³ ³    exponent digits
;     bits: 7654³3210                       ³ ³
;           ÂÂXXÁÂÂÂÂ                     E E EÄÄÄ Exponent
;           ³³³³ ³³³³                     ³
;           ³³³³ ³³³³                     ³ 
;           ³³³³ ÀÁÁÁÄÄÄÄÄÄÄÄÄÄÄÄÄÄÄÄÄÄÄÄÄÁÄÄÄÄÄÄ- High exponent
;           ³³³³                                   digit
;           ³³ÀÁÄÄ Error reporting flags
;           ³ÀÄÄÄÄ Exponent Sign
;           ÀÄÄÄÄÄ Mantissa Sign
;
;
;   Any number represented in this format can also be represented in
;   the following display format:
;
;           (ñSm)D.DDDDDDDDDDD(ñSe)EEE
;
;   where the 'D's stand for a decimal mantissa digit and the 'E's
;   stand for decimal exponent digits. 'Sm' stands for a single bit
;   mantissa sign ( 0 = +, 1 = - ), and 'Se' stands the sign of the
;   exponent.
;
;   A convention observed in the following code is that a test for zero
;   can be done by simply testing the third byte of the number for
;   zero. This works because for any non-zero number the mantissa is
;   left justified until its two most significant digits are in this
;   location. Only zero (or an equivalent caused by underflow) will
;   have this location empty.
;
;   There are two unused bits in this storage format. Bits 4 and 5 of
;   the first byte (see above) are not assigned. One possible assignment
;   strategy uses these bits to identify invalid arguments. In this
;   version we have begun to implement it. We use the following format
;   for byte 0:
;
;                    x x     x x x x
;           byte 0:  - - 5 4 - - - -
;
;                        0 0            Valid number
;                        0 1            Underflow (not reported)
;                        1 0            Overflow
;                        1 1            Signal  (domain, range, ... error)
;
;   where the 'signal' is expressed by a BCD number in the exponent
;   position. When an invalid argument is detected, the
;   called routine will return with one of the following 8-byte strings:
;
;           $29,$99,$99,$99,$99,$99,$99,$99 : Overflow error
;           $30,1,0,0,0,0,0,0               : Range error
;           $30,2,0,0,0,0,0,0               : Domain error
;           $30,3,0,0,0,0,0,0               : Divide-by-zero error
;
;   The order of the two arguments in binary operations is given by:
;
;
;   add:    w1 + w2 -> w3
;   sub:    w1 - w2 -> w3
;   mul:    w1 * w2 -> w3
;   div:    w1 / w2 -> w3
;   pow:    w1 ^ w2 -> w3
;   nCr:    w1 C w2 -> w3
;   nPr:    w1 P w2 -> w3
;
;   For unary operations, the argument is in 'w1' and the result in 'w3':
;
;   sqrt:      w1   -> w3
;   int:       w1   -> w3
;   frac:      w1   -> w3
;   abs:       w1   -> w3
;   sine:      w1   -> w3
;    ...
;

;-------------------------------------------------------------------------
;
; Variables used by the core math routines
;

tmp1	=	$C0	; tmp1, tmp2 : 2 bytes each, also used as 16-bit tmp's
tmp2	=	$C2
dadj	=	$C4
ctr1	=	$C5
ctr2	=	$C6
ctr3	=	$C7
pctr	=	$C8
bctr	=	$C9
tnsgn	=	$CA
qdrnt	=	$CB
cotflg	=	$CC
cosflg	=	$CD
sinflg	=	$CE
expflg	=	$CF
cdiff	=	$D0	; set if diffexp exits with C=1 (e1 >= e2)
eqexp	=	$D1	; set if e1 and e2 have same absolute values after diffexp


;-------------------------------------------------------------------------
; Pointers used by various routines

mptr	=	$F0		; Used when building multiples table
ptr1	=	$F2
ptr2	=	$F4
ptr3	=	$F6


; ------------------------------------------------------------------------
;
; other RAM assignments

mtbl	=	$0200	; tables of multiples (0..9)
mtbl1	=	$0210	; must begin at a page boundary
mtbl2	=	$0220	; from $0200 to $029F

ra	= $0300		; These registers are for mantissa operations.
rb	= $0310
rc	= $0320
rd	= $0330
re	= $0340
rf	= $0350

w1	= $0360		; These registers are for input of numbers in
w2	= $0368		; storage format. 'w1' and/or 'w2' should be
w3	= $0370		; loaded with the operands before calling
w4	= $0378		; F.P. routines.  'w3' is the result.
w5	= $0380		; w5 is used by new functions

scalereg = $0388	; register for scaling parameter
arg	 = $0390	; argument register for poly routine
reg1	 = $0398	; temporary storage registers
reg2	 = $03A0
reg3	 = $03A8
reg4	 = $03B0


st_L	= $03B8		; RPN user stack - LASTx register
st_X	= $03C0		; RPN stack - X register
st_Y	= $03C8		; RPN stack - Y register
st_Z	= $03D0		; RPN stack - Z register
st_T	= $03D8		; RPN stack - T register


;-------------------------------------------------------------------------
;
;   misc equates
;

reglen	= 8		; register length in bytes
wreglen	= 16		; register length if in 16-bit words


; -------------------------------------------------------------
;
;   main floating point math package
;

sub	lda	w2	; perform subtraction by changing sign
	eor	#$80	; of second number and adding.
	sta	w2
add	sed
	lda	w1+2	; filter out case where one number is zero.
	bne	add1
	lda	w2+2	; w1 is zero, so if w2 is also zero...
	bne	add4
	stz	w2	; ...reset its sign
add4	lda	#<w2
	ldy	#>w2
	ldx	#<w3	; this convention references the low byte.
	jmp	copy2w	; w1 is zero -- return with w2 (which may also be zero).
add1	lda	w2+2
	bne	add2
	lda	#<w1
	ldy	#>w1
	ldx	#<w3
	jmp	copy2w	; w2 is zero -- return with w1 (which won't be zero).
    ;   At this point, additions and subtractions involving '0.0' have
    ;   been handled. Both operands are now known to be non-zero.
    ;
add2	lda	w1	; determine if a sum or difference is required.
	eor	w2
	bpl	add3	; branch if signs are the same, drop through otherwise
    ;
    ;  On entry, we only know that the two numbers have opposite
    ;  signs (mantissa signs). To subtract, we will determine
    ;  the larger of the two numbers (in magnitude) and place it
    ;  in 'ra'. The second number will be decimal aligned and
    ;  placed in 'rb'.
    ;
    ;  The sign of the result will carry the sign of the number
    ;  in 'ra' (largest in magnitude).
    ;
sub1	bit	w1	; now estimate which number is larger from tests
	bvs	est1	;  on the exponents.  If exponent signs are unequal,
	bit	w2	;  the positive one goes with the larger magnitude.
	bvs	est2
	bvc	sbpe	; both numbers have positive exponents.
est1	bit	w2
	bvs	sbne	; both numbers have negative exponents.
	jsr	xw12
est2	jsr	sumexp	; exponents have opposite signs -- e1 is larger.
	lda	tmp1
	beq	est3
est2a	lda	#<w1	; e1 >> e2
	ldy	#>w1	; just copy e1 to w3...
	ldx	#<w3
	jmp	copy2w	; ...and exit.
est3	lda	tmp2
	cmp	#$12
	bcs	est2a	; e1 >> e2
			; we now know that a subtraction will take place.
	stz	dadj	; initialize decimal adjust.
	jsr	w12rab	; put 'w1' and 'w2' into 'ra' and 'rb' respectively.
	jsr	subrarb
	bcs	est4a
	jsr	xw12

	stz	dadj
	jsr	w12rab
	jsr	subrarb
est4a	lda	w1	; 'tmp1' is used as a flag to identify the caller...
	eor	#$40
	sta	tmp1
	jmp	rbtow3	; ...of this subroutine.
			; jsr, rts

sbpe	jsr	diffexp	; carry e1 >= e2.
	bcs	sbp1
	jsr	xw12
sbp1	lda	tmp1
	bne	est2a
	bra	est3
sbne	jsr	diffexp	; exponent diff: tmp1(hi),tmp2(lo), carry: e1 >= e2.
	bcc	sbp1
	jsr	xw12
	lda	tmp1
	bra	sbp1
    ;
    ;   We now have two numbers with the same sign. 
    ;   Add two positive or negative numbers.
    ;   Handle sign on exit, as it does not affect the addition.
    ;
    ;   The following algorithm places the larger of the two
    ;   numbers in 'ra' and the other in 'rb', but only if a
    ;   decimal point adjustment is needed before the addition.
    ;
add3	bit	w1	; now estimate which number is larger from tests
	bvs	et1	;  on the exponents.  If exponent signs are unequal,
	bit	w2	;  the positive one goes with the larger magnitude.
	bvs	et2
	bvc	adpe	; both numbers have positive exponents.
et1	bit	w2
	bvs	adne	; both numbers have negative exponents.
	jsr	xw12	; swap w1 and w2 so largest magnitude is in w1.
et2	jsr	sumexp	; add the exponents to determine the decimal
			;  point adjustment to be applied to w2.
	lda	tmp1	; if tmp1 <> 0 the sum exceeds number of digits
	beq	et3
et2a	lda	#<w1
	ldy	#>w1	; w1 >> w2. no need to add, just copy to w3...
	ldx	#<w3
	jmp	copy2w	; ...and exit
et3	lda	tmp2
	cmp	#$12
	bcs	et2a	; w1 >> w2. copy and exit.
			; Summary: w1 is larger than w2, decimal adj in tmp2.
	stz	dadj	; initialize adjustments due to potential overflow.
	jsr	w12rab	; copy w1 and w2 to working regs with decimal adj.
	jsr	addrarb	; perform the addition with overflow adjustment.
	jsr	rndrb	; (optional) round result register.
	lda	w1          
	sta	tmp1	; 'tmp1' identifies caller to next subroutine.
	jmp	rbtow3	; copy result (rb) to w3.
			; jsr, rts

adpe	jsr	diffexp	; exponent diff: 'a'(hi), 'x'(lo), carry: e1 >= e2.
	bcs	adp1
	jsr	xw12
adp1	lda	tmp1
	bne	et2a	; e1 >> e2, just copy w1 to w3 and exit.
	bra	et3
adne	jsr	diffexp	; exponent diff: tmp1(hi), tmp2(lo), carry: e1 >= e2.
	bcc	adp1
	jsr	xw12
	bra	adp1


sumexp	lda	w1	; return with sum of exponents in tmp1(hi),tmp2(lo)
	and	#$0f
	sta	tmp1
	lda	w2
	and	#$0f
	sta	tmp2
	clc
	lda	w1+1
	adc	w2+1
	tax
	lda	tmp1
	adc	tmp2
	sta	tmp1
	stx	tmp2
	rts


diffexp	lda	w1	; return absolute difference of exponents of w1 and w2.
	and	#$0f	; value is in tmp1(hi),tmp2(lo). carry set if e1 >= e2,
	sta	tmp1	;  clear if e1 < e2.
	lda	w2
	and	#$0f
	sta	tmp2
	sec		; trial difference assuming e1 >= e2
	lda	w1+1	; (It's just as easy to go ahead and perform the
	sbc	w2+1	;  subtraction as it is to perform a multibyte
	tax		;  compare. If the numbers were properly ordered,
	lda	tmp1	;  we are done. Otherwise, we subtract in reverse
	sbc	tmp2	;  order, which we would have done anyway.)
	bcs	lab001	; OK, exit with carry set.
	sec		; Underflow. e1 is smaller than e2, so recalculate
	lda	w2+1	;  exponent difference.
	sbc	w1+1
	tax
	lda	tmp2
	sbc	tmp1
	clc
lab001	sta	tmp1
	stx	tmp2
	rts


    ;
    ;   Copy mantissas of numbers in 'w1' and 'w2' to 'ra' and 'rb'.
    ;   Align decimal points by shifting 'rb' to the right. Decimal
    ;   adjustment has been previously placed in 'tmp2'.
    ;   

w12rab	ldy	#wreglen*2-1	;31
	lda	#0
clrlp	sta	ra,y	; clear ra, rb
	dey
	bpl	clrlp
	clc		; determine register alignment offset.
	lda	tmp2
	adc	#$11	; this 'magic' number is 2x5+1.
	tay
	lda	dec2bin,y	; convert to binary...
	lsr		; ... then convert to bytes
	tax		; decimal point adjustment is now contained in 'x'
	ldy	#reglen-3
lab002	lda	w1+2,y	; now copy w1 to ra and w2 to rb (mantissas only)
	sta	ra+2,y
	lda	w2+2,y
	sta	rb+2,x
	dex
	dey
	bpl	lab002
	lda	tmp2
	lsr		; set carry if decimal adjustment is odd...
	bcc	cxp1
	ldx	#wreglen-3	; ...and shift rb left one digit to normalize.
	lda	#0
lab003	ldy	rb+2,x
	ora	tlh,y
	sta	rb+2,x
	lda	thl,y
	dex
	bpl	lab003
cxp1	rts


addrarb	ldx	#reglen-2	; add previously aligned 'ra' to 'rb', result in 'rb'
	clc
lab005	lda	ra+1,x
	adc	rb+1,x
	sta	rb+1,x
	dex
	bpl	lab005
	lda	rb+1	; check for overflow...
	beq	lab006
	jsr	rotrgtb	; ...and normalize, if needed
	inc	dadj	; and increment decimal adjustment.
lab006	rts

subrarb	ldx	#reglen
	sec
lab007	lda	ra+1,x
	sbc	rb+1,x
	sta	rb+1,x
	dex
	bpl	lab007
	bcc	sabx	; exit with carry clear (underflow condition).
	lda	#12
	sta	tmp1
sab1	lda	rb+2	; normalize
	and	#$f0
	bne	sabxaa
	jsr	shflftb
	inc	dadj	; this quantity will be subtracted from the exponent.
	dec	tmp1
	bne	sab1
sabxaa	ldy	dadj	; convert dadj to decimal
	lda	bin2dec,y
	sta	dadj
	sec		; exit with carry set (normal condition).
sabx	rts

rndrb	ldx	#7	; round the result in 'rb'
	lda	#$50
	clc
lab008	adc	rb+1,x
	sta	rb+1,x
	lda	#0
	dex
	bpl	lab008
	lda	rb+1	; check for overflow...
	beq	lab009
	jsr	rotrgtb	; ...and normalize, if needed...
	clc
	lda	dadj	; assume dadj < 99 but (possibly) > 9
	adc	#1
	sta	dadj	; ... then increment decimal adjustment.
lab009	rts

rbtow3	ldx	#5	; copy rb to w3.
	clc
rbtw1	lda	rb+2,x	; first, copy result mantissa from rb to w3.
	sta	w3+2,x
	dex
	bpl	rbtw1
	lda	w3+2
	bne	rbta	; check if zero...
	sta	w3+1	; ...and exit if so.
	sta	w3
	rts

rbta	lda	w1+1	; now copy exponent from w1 to w3
	sta	w3+1
	lda	w1
	and	#$0f
	sta	w3
	bit	tmp1	; adjust exponent, assume dadj < 99
	bvc	rbtw2	; add adjustment to exponent
	sec
	lda	w3+1
	sbc	dadj
	tax
	lda	w3
	sbc	#0
	bcs	rbtxa
	sec
	lda	dadj
	sbc	w3+1
	sta	w3+1
	lda	#0
	sbc	w3
	sta	w3
	lda	w1
	and	#$c0
	eor	#$40
	ora	w3
	and	#$cf
	sta	w3
	rts

rbtw2	clc
	lda	w1+1
	adc	dadj
	tax
	lda	w1
	adc	#0
rbtxa	sta	w3	; save exponent
	stx	w3+1
	ora	w3+1
	bne	rbtx
	lda	w1	; if exponent is zero...
	and	#$80	; ...preserve mantissa sign, suppress exponent sign.
	ora	w3
	and	#$cf
	sta	w3
	rts

rbtx	lda	w1
	and	#$c0	; now copy sign bits to w3.
	ora	w3
	and	#$cf
	sta	w3
	rts

xw12	ldy	#7	; exchange contents of w1 with w2.
lab010	ldx	w1,y
	lda	w2,y
	sta	w1,y
	txa
	sta	w2,y
	dey
	bpl	lab010
	rts

shflftb	ldx	#7	; shift mantissa in rb left one digit.
	lda	#0
lab011	ldy	rb+1,x
	ora	tlh,y
	sta	rb+1,x
	lda	thl,y
	dex
	bpl	lab011
	rts

rotrgtb	ldx	#7
	lda	#0
lab012	ldy	rb+1,x
	ora	tlh,y
	sta	rb+2,x
	lda	thl,y
	dex
	bpl	lab012
	lda	#0
	sta	rb+1
	rts


;
;   multiply routines
;
;   Multiplication is performed by a novel algorithm involving a
;   table of multiples which is pre-computed for each call.
;   Instead of repeatedly adding suitably aligned versions of the
;   multiplier to an accumulated sum, a table consisting of
;   multiples of the multiplicand is accessed by indexing on the
;   current multiplier digit.
;
;   The rationale behind the method is that there are basically
;   two major repetitive operations involved in multiplication.
;   One is the addition of one multi-digit register to another,
;   and the other is a single digit shift of a multi-digit
;   register.
;
;   For 8-bit processors using multi-digit packed BCD numbers
;   (two digits per byte) a two digit shift can be done by
;   simply indexing the bytes by 1 (1 byte equals 2 digits).
;
;   Minimizing the number of BCD adds and shifts is the driving
;   force behind the choice of this algorithm.
;
;   The fixed point table for a specific case looks as follows:
;
;   Multiplicand = 23.456...
;
;       Table
;
;   0   0000000...
;   1   0234560...
;   2   0469120...
;   3   0703680...
;   4   0938240...
;   5   1172800...
;   6   1407360...
;   7   1641920...
;   8   1876480...
;   9   2111049...
;
;   The partial product is updated in two phases. All digits on even
;   boundaries are processed in one pass, and the digits on odd boundaries
;   are processed in the second. This improves the performance of BCD
;   computations by eliminating all but one multi-byte register shift.
;   Note that indexing through memory a byte at a time skips 2 BCD
;   digits at a time.
;
;   Suppose the multiplier is:
;
;               1234567890
;
;   Then the partial products are accumulated during the first pass as if
;   the multiplier were:
;
;               0103050709
;
;   Since the non-zero digits are on byte boundaries, the register
;   alignment can be handled by simply incrementing or decrementing
;   the value in an index register. It is not necessary to shift the
;   multi-digit register during this phase.
;
;   The result register is then shifted left one digit, and the second
;   pass accumulates partial products as if the multiplier were:
;
;               0204060800
;
;
;   The algorithm is summarized as follows:
;
;       1. Create the table of multiples of the multiplicand,
;       2. For each digit of the multiplier on even boundaries:
;           A. get next digit of multiplier,
;           B. using digit as an index, access corresponding table entry,
;           C. add table entry to partial product with proper
;               byte alignment,
;           D. if more digits, goto A.
;       3. Shift multiplier one digit,
;       4. Repeat (2.) for remainder of digits.
;
;   To estimate the performance of this strategy, consider a typical
;   repeated addition algorithm for multi-digit BCD numbers. On the
;   average, there will be 4.5 multi-byte additions per digit of the
;   multiplier.
;   After each digit is processed, the partial product register must
;   be shifted one digit. For 10 digit numbers, there will be
;   approximately 45 adds and 9 register shifts.
;
;   For the method used here, there are 8 adds required to construct
;   the table, and one add per digit of the multiplier. There is only
;   one register shift required, so the total operation count will be
;   18 adds and 1 register shift.
;
;   12 / 2014 : rewritten main routine - M. Ferrer
;               fixed exponents sidecases
;               (screwed up on multiplying 2 numbers with exponents
;               having same absolute values but opposite signs)
;
;   09 / 2017 : add test w1 = 0 or w2 = 0  ->  return zero without
;               doing any computation
;               (before that, we got a result with a computed zero
;               mantissa, and a non-zero exponent field)
;

mul	lda	w1+2	; weed out case w1 = 0 or w2 = 0
	beq	mul01

	lda	w2+2
	bne	mul02

mul01	jmp	retzr	; one of them = 0 : return 0 in w3
			; jsr, rts
mul02	sed
	stz	dadj	; reset exponent adjust flag
	stz	cdiff	; will be 1 if C=1 from diffexp
	stz	eqexp	; will be 1 if abs(e1) = abs(e2)
	jsr	rr_clr	; clear working registers

	jsr	mulm	; get product of mantissas
	lda	w1	; check exponents for same or different signs
	eor	w2
	and	#$40
	bne	lab014

	jsr	sumexp	; same signs, sum of expnts in tmp1/tmp2
	lda	w1	; copy sign in w3
	and	#$40
	sta	w3
	lda	dadj	; exponent adjustment needed ?
	beq	cxpn

	lda	w1	; yes, test exponent sign
	and	#$40
	bne	s8xpn	; negative, decrement its absolute valuie

	jsr	incexp	; positive, increment its absolute value
	bra	cxpn

s8xpn	jsr	decexp
	bra	cxpn

lab014	jsr	diffexp	; different exponent signs
	php		; save result
	lda	tmp1	; test abs(e1) and abs(e2) equality
	eor	tmp2
	bne	s6xpn
	dec	eqexp	; same absolute values

s6xpn	plp		
	bcc	s0xpn
	dec	cdiff	; exited with C=1 : e1 >= e2

s0xpn	lda	dadj	; exponent adjustment needed ?
	beq	s2xpn

	lda	cdiff	; yes, choose if +1 or -1
	and	#$40
	eor	w1
	and	#$40
	beq	s1xpn

s5xpn	jsr	incexp
	bra	s2xpn

s1xpn	lda	eqexp	; if e1 = e2 special case -> incexp
	bne	s5xpn
	jsr	decexp

s2xpn	lda	eqexp	; sign of w3 exponent
	beq	s7xpn
	lda	#0	; positive if e1 = e2
	sta	w3
	bra	cxpn

s7xpn	lda	cdiff	; else ...
	beq	s3xpn
	lda	w1	; is w1's if e1 > e2
	bra	s4xpn

s3xpn	lda	w2	; is w2's if e1 < e2
s4xpn	and	#$40
	sta	w3

cxpn	lda	tmp2	; tmp1/tmp2 to exponent field of w3
	sta	w3+1
	ora	tmp1	; if exponent is zero, clear exponent sign
	bne	cxp2

	sta	w3
cxp2	lda	tmp1
	ora	w3	; add exponent sign
	sta	w3
	lda	w1	; now correct mantissa sign
	eor	w2
	and	#$80
	ora	w3
	sta	w3	; add mantissa sign
	rts


mulm	lda	w1+2	; multiply mantissas
	beq	mzro
	lda	w2+2
	beq	mzro
	jsr	cw12rab
	jsr	mktbl	; create table of multiples...
	jsr	psumh	; ...accumulate partial product...
	jsr	shflft	; ...shift...
	jsr	psuml	; ...now accumulate rest of partial product.
	jmp	rc2w3	; jsr, rts

mzro	ldx	#7
lab015	lda	#0
	sta	w3,x
	dex
	bpl	lab015
	rts

mktbl	lda	#<mtbl	; create table of multiples of multiplier
	sta	mptr
	lda	#>mtbl
	sta	mptr+1
	sta	ptr1+1
	sta	ptr2+1
	lda	#$10
	sta	ptr1
	lda	#$20
	sta	ptr2
	ldx	#7
lab016	lda	ra+9,x	; copy multiplier to 1st multiple     
	sta	mtbl+$10,x
	dex
	bpl	lab016
	ldx	#8
mkolp	ldy	#7	; now compute all multiples
	clc
lab017	lda	(ptr1),y
	adc	mtbl1,y	; BCD add
	sta	(ptr2),y
	dey
	bpl	lab017
	cld
	clc		; bump pointers for next entry
	lda	ptr1
	adc	#$10
	sta	ptr1
	clc
	lda	ptr2
	adc	#$10
	sta	ptr2
	sed
	dex
	bne	mkolp
	rts

cw12rab	ldx	#5
lab018	lda	w1+2,x
	sta	ra+10,x
	lda	w2+2,x
	sta	rb+10,x
	dex
	bpl	lab018
	rts
;
;   1st partial sum. add sums of products of high digits from
;   each multiplier byte.
;
psumh	lda	#>ra	; set up pointers to working registers
	sta	ptr1+1
	sta	ptr2+1
	lda	#$18
	sta	ptr1
	lda	#$28
	sta	ptr2
	ldy	#7
psh1	lda	(ptr1),y	; get next high digit
	and	#$f0
	beq	psh2	; if zero, no need to add
	sta	mptr	; update pointer to multiple table for this digit
psh1a	lda	(mptr),y	; copy this multiple to result register
	sta	(ptr2),y
	dey
	bpl	psh1a
psh2	ldx	#6	; index for number of additional digits in this pass
molp	dec	ptr1	; bump pointers down one byte
	dec	ptr2
	ldy	#7	; index for register length
	lda	(ptr1),y	; get next digit
	and	#$f0
	beq	noadd	; if zero, no need to add
	sta	mptr
	clc
lab019	lda	(mptr),y	; add multiple of current digit to partial sum
	adc	(ptr2),y
	sta	(ptr2),y
	dey
	bpl	lab019
noadd	dex
	bpl	molp	; next digit
	rts
;
;   2nd partial sum. add sums of products with low digits from
;   each multiplier byte.
;
psuml	lda	#$19
	sta	ptr1
	lda	#$29
	sta	ptr2
	ldx	#6
psl1	dec	ptr1
	dec	ptr2
	ldy	#7
	lda	(ptr1),y
	and	#$0f
	beq	psl3	; if zero, no need to add
	asl
	asl
	asl
	asl
	sta	mptr	; update pointer to table of multiples for this byte
	clc
psl2	lda	(mptr),y
	adc	(ptr2),y
	sta	(ptr2),y
	dey
	bpl	psl2
	bcc	psl3
	ldy	ptr2
	dey
lab020	lda	ra,y
	adc	#0
	sta	ra,y
	dey
	bcs	lab020
psl3	dex
	bpl	psl1
	rts

shflft	ldx	#15	; shift 'rc' left
	lda	#0
lab021	ldy	rc,x
	ora	tlh,y
	sta	rc,x
	lda	thl,y
	dex
	bpl	lab021
	rts

rc2w3	lda	#1	; copy/align 'rc' to 'w3'
	sta	dadj
	lda	rc+3
	and	#$f0
	bne	rct1
	jsr	shflftc
	dec	dadj
rct1	ldx	#5
lab022	lda	rc+3,x
	sta	w3+2,x
	dex
	bpl	lab022
	rts

shflftc	ldx	#7	; shift mantissa in 'rc' left one digit.
	lda	#0
lab023	ldy	rc+2,x
	ora	tlh,y
	sta	rc+2,x
	lda	thl,y
	dex
	bpl	lab023
	rts


incexp	clc		; increment an exponent in tmp1(H)/tmp2(L)
	lda	tmp2
	adc	#1
	sta	tmp2
	lda	tmp1
	adc	#0
	sta	tmp1
	rts


decexp	sec		; decrement an exponent in tmp1(H)/tmp2(L) 
	lda	tmp2
	sbc	#1
	sta	tmp2
	lda	tmp1
	sbc	#0
	sta	tmp1
	rts


;
;   divide routines
;
;   The divide algorithm uses tables of multiples, as does the
;   multiply algorithm. In this case each multiple is a candidate
;   for subtraction from the current remainder. A quick scan
;   weeds out multiples which are unsuitable for the trial
;   subtraction and a recovery routine is invoked in the event
;   that an underflow occurs anyway.
;
div	sed		; weed out zero divisor or dividend
	lda	w2+2
	bne	lab024

	jmp	dvzrerr	; divisor is zero, signal error

lab024	lda	w1+2
	bne	d1b

	jmp	retzr	; dividend is zero, return 0.00--e000
			; jsr, rts
d1b	lda	#0
	ldx	#47
d1	sta	ra,x	; clear out working area
	dex
	bpl	d1
	ldx	#5	; copy mantissas from w1/w2 to ra/rb
d2	lda	w2+2,x	; (ra is divisor, rb is working reg/quotient)
	sta	ra+10,x
	lda	w1+2,x
	sta	rb+2,x
	dex
	bpl	d2
	jsr	mktbl	; create table of multiples
	jsr	shftlft	; initialize remainder for 1st digit divide
	lda	#15	; digit counter
	sta	ctr2
onedig	jsr	decdiv	; divide, one digit at a time
	dec	ctr2
	bne	onedig
	lda	#1	; copy result to w3 mantissa with dadj = exponent adj.
	sta	dadj
	lda	rb+8
	beq	od1
	jsr	shftlft
	dec	dadj
	ldx	#5
od2	lda	rb+8,x
	sta	w3+2,x
	dex
	bpl	od2
	bmi	od4	; branch always
od1	ldx	#5
od3	lda	rb+9,x
	sta	w3+2,x
	dex
	bpl	od3
;
;   handle exponent
;
od4	lda	w1
	eor	w2
	and	#$40	; compare exponent signs
	beq	exd
	jsr	sumexp
	lda	w1
	bra	dexpadj
exd	jsr	diffexp	; exponent signs are the same
	lda	w2
	bcs	dexpadj	; carry: divisor exp >= dividend exp (e1 >= e2)
	eor	#$40
dexpadj	and	#$40
	sta	w3
	bit	w3
	bvs	od5	; branch to negative exponent handler
	lda	dadj	; positive exponent
	beq	od6
	lda	tmp2
	ora	tmp1
	bne	od5a
	lda	#1	; unadjusted exponent is zero and decrement is needed
	sta	tmp2
	lda	#$40
	ora	w3
	sta	w3
	bra	od6
od5	lda	dadj
	beq	od6
	jsr	incexp
	bra	od6
od5a	jsr	decexp
od6	lda	tmp2	; now copy adjusted exponent
	sta	w3+1
	lda	tmp1
	ora	w3
	sta	w3
	and	#$0f	; test for condition of zero exponent...
	ora	w3+1	; ...and clear exponent sign, if so
	bne	odx
	sta	w3
odx	lda	w1	; finally, set +/- mantissa sign.
	eor	w2
	and	#$80
	ora	w3
	sta	w3
	rts

decdiv	lda	#9	; compute next quotient digit
	sta	ctr1
	lda	rb
	ldx	rb+1
	cmp	mtbl+$90	; check 9th multiple for valid subtract
	bcc	lab025
	bne	dsub
	cpx	mtbl+$91
	bcs	dsub
lab025	dec	ctr1
	cmp	mtbl+$80	; check 8th multiple
	bcc	lab026
	bne	dsub
	cpx	mtbl+$81
	bcs	dsub
lab026	dec	ctr1
	cmp	mtbl+$70	; check 7th multiple
	bcc	lab027
	bne	dsub
	cpx	mtbl+$71
	bcs	dsub
lab027	dec	ctr1
	cmp	mtbl+$60	; check 6th multiple
	bcc	lab028
	bne	dsub
	cpx	mtbl+$61
	bcs	dsub
lab028	dec	ctr1
	cmp	mtbl+$50	; check 5th multiple
	bcc	lab029
	bne	dsub
	cpx	mtbl+$51
	bcs	dsub
lab029	dec	ctr1        
	cmp	mtbl+$40	; check 4th multiple
	bcc	lab030
	bne	dsub
	cpx	mtbl+$41
	bcs	dsub
lab030	dec	ctr1
	cmp	mtbl+$30	; check 3rd multiple
	bcc	lab031
	bne	dsub
	cpx	mtbl+$31
	bcs	dsub
lab031	dec	ctr1
	cmp	mtbl+$20	; check 2nd multiple
	bcc	lab032
	bne	dsub
	cpx	mtbl+$21
	bcs	dsub
lab032	dec	ctr1
	cmp	mtbl+$10	; check 1st multiple    
	bcc	lab033
	bne	dsub
	cpx	mtbl+$11
	bcs	dsub
lab033	dec	ctr1	; must be zero
;	and	rb+15,$f0
	and	rb+15
	jmp	shftlft
dsub	lda	rb+15	; save current digit in result register
	and	#$f0
	ora	ctr1
	sta	rb+15
	asl
	asl
	asl
	asl
	sta	ptr1
	ldy	#7
	sec
sulp	lda	rb,y	; shift remainder (and quotient) left one digit
	sbc	(ptr1),y
	sta	rb,y
	dey
	bpl	sulp
	bcs	shftlft	; branch if no correction is needed
	dec	rb+15	; here we recover from underflow caused by
	ldy	#7	;  subtracting a too-large multiple
lab034	lda	rb,y	; correct remainder after underflow
	adc	mtbl+$10,y	;  by adding 1st multiple back in
	sta	rb,y
	dey
	bpl	lab034
shftlft	ldx	#15	; shift remainder register in rb left one digit.
	lda	#0
lab035	ldy	rb,x
	ora	tlh,y
	sta	rb,x
	lda	thl,y
	dex
	bpl	lab035
	rts                         


;
;   square root routines
;
;   (C) 1999, C. Bond. Finds square root of BCD number by
;       non-restoring pseudo-division.
;
sqrt	sed
	lda	w1+2
	bne	sq0	; check for zero

	jmp	retzr	; square root of zero, clear result and return
			; jsr, rts

sq0	bit	w1	; check for negative number...
	bpl	sq0b	; ...and return with domain error.
	jmp	domnerr
sq0b	ldx	#15
	lda	#0
lab037	sta	ra,x	; clear working register
	dex
	bpl	lab037
	ldx	#7
lab038	lda	w1+2,x
	sta	ra+2,x	; now copy mantissa to ra
	dex
	bpl	lab038
	lda	w1+1	; check for odd exponent to force grouping of digits
	and	#1
	bne	sq2
	ldx	#7	; even exponent, shift right one digit
	lda	#0
lab039	ldy	ra,x
	ora	tlh,y
	sta	ra+1,x
	lda	thl,y
	dex
	bpl	lab039
sq2	ldx	#7	; multiply mantissa by 5
	clc
lab040	lda	ra+1,x
	adc	ra+1,x
	sta	rb+1,x
	dex
	bpl	lab040
	ldx	#7	; rb = 2 * ra
	clc
lab041	lda	rb+1,x
	adc	rb+1,x
	sta	rb+1,x
	dex
	bpl	lab041
	ldx	#7
	clc
lab042	lda	ra+1,x
	adc	rb+1,x
	sta	ra+1,x
	dex
	bpl	lab042
	lda	#1	; ra = 5 * ra
	sta	ptr1	; index used by pseudo-divisor
	stz	rb
	lda	#6	; number of digits counter (x2)
	sta	ctr1
sq2d	ldx	ptr1
	lda	#$05
	sta	rb,x
	jsr	sq3
	dec	ctr1
	bmi	sqx
	jsr	sarlft
	bra	sq2d
sqx	ldx	#7
	lda	#0
lab043	ldy	rb+1,x	; shift rb left one digit
	ora	tlh,y
	sta	rb+1,x
	lda	thl,y
	dex
	bpl	lab043
	ldx	#5
sqx1b	lda	rb+1,x	; move mantissa to result register
	sta	w3+2,x
	dex
	bpl	sqx1b
	lda	w1+1	; calculate exponent (divide by 2)
	sta	w3+1
	lda	w1
	and	#$0f
	sta	w3
	lsr	w3
	ror	w3+1	; shift exponent right one bit...
	ldy	w3
	lda	srtbl,y	; ...and translate to perform BCD correction.
	sta	w3
	ldy	w3+1
	lda	srtbl,y
	sta	w3+1
	bit	w1	; is exponent negative?
	bvc	sqxx
	lda	w1+1	; if so, adjust the exponent... 
	and	#$1
	beq	sqxc
	clc
	adc	w3+1
	sta	w3+1
	lda	w3
	adc	#0
	sta	w3
sqxc	lda	#$40	; ...and correct the result sign
	ora	w3
	sta	w3
sqxx	rts

sq3	ldx	ptr1
	sec
lab044	lda	ra,x
	sbc	rb,x
	sta	ra,x
	dex
	bpl	lab044
	bcc	sq4
	ldx	ptr1
	clc
	lda	rb,x
	adc	#$10
	sta	rb,x
	bcc	sq3
sq4	ldx	ptr1
	lda	rb,x
	and	#$f0
	ora	#$9
	sta	rb,x
	lda	#$50
	sta	rb+1,x
	jsr	sarlft
	inc	ptr1
sq5	ldx	ptr1
	clc
lab045	lda	ra,x
	adc	rb,x
	sta	ra,x
	dex
	bpl	lab045
	bcs	sq6
	ldx	ptr1
	dec	rb-1,x
	bcc	sq5
sq6	rts

sarlft	ldx	#15
	lda	#0
lab046	ldy	ra,x
	ora	tlh,y
	sta	ra,x
	lda	thl,y
	dex
	bpl	lab046
	rts


;
;   secant -- calculates secant function: sec(x) = sqrt(1 + tan^2(x))
;
;   w3 = secant(w1)
;
secnt	jsr	tan
	lda	#<w3
	ldy	#>w3
	ldx	#<w4
	jsr	copy2w	; save tangent for (possible) later use
	lda	#<w3
	ldy	#>w3
	ldx	#<w1
	jsr	copy2w
	lda	#<w3
	ldy	#>w3
	ldx	#<w2
	jsr	copy2w
	jsr	mul	; w3 contains tan^2
	lda	#<w3
	ldy	#>w3
	ldx	#<w2
	jsr	copy2w
	lda	#<unit
	ldy	#>unit
	ldx	#<w1
	jsr	copy2w	; w3 = 1 + tan^2
	jsr	add
	lda	#<w3
	ldy	#>w3
	ldx	#<w1
	jsr	copy2w
	jsr	sqrt	; w3 = sqrt(1 + tan^2), w4 = tan
	lda	qdrnt
	cmp	#1
	beq	lab047
	cmp	#4
	beq	lab047
	lda	w3
	ora	#$80
	sta	w3
lab047	rts


;
;   sin -- calculates sine function:  sin(x) = tan(x)/sec(x)
;
sin	jsr	secnt
	lda	#<w3
	ldy	#>w3
	ldx	#<w2
	jsr	copy2w
	lda	#<w4
	ldy	#>w4
	ldx	#<w1
	jsr	copy2w
	jsr	div
	lda	w3
	and	#$7f
	ldx	qdrnt
	cpx	#3
	bcc	lab048
	ora	#$80
lab048	sta	w3
	rts


;
;   cos -- cos(x) = 1/sec(x)
;
cos	jsr	secnt
	lda	#<w3
	ldy	#>w3
	ldx	#<w2
	jsr	copy2w
	lda	#<unit
	ldy	#>unit
	ldx	#<w1
	jsr	copy2w
	jmp	div	; jsr, rts


;
;   CORDIC routine to calculate tangent of given angle (radians)
;
;   On entry, 'w1' contains the angle, on exit 'w3' contains the tangent.
;
;   'ra' is used for resolving the angle, where 're' accumulates the
;   pseudo-quotient. After this 'ra', 'rb', 'rc' and 'rd' are the
;   working registers for pseudo-multiplication.
;
;   The computation strategy is as follows:
;
;   1. Scale the given argument.
;      . divide the argument by 2pi,
;      . discard the integer part of the result,
;      . multiply the fractional part by 2pi.
;   2. If the result < 0 add 2pi so reference angle is positive.
;   3. Find and set quadrant and tnsgn, quadrant is used by sin, cos, etc.
;   4. If arg > pi subtract pi
;   5. if arg > pi/2 subtract pi
;   6. The scaled argument has the same tangent as the original argument.
;      It is now in quadrant I or IV.
;      . take the absolute value of the argument.
;   7. The argument is now in the range 0 <= arg <= pi/2.
;      . set 'cotflg' = 0
;   8. If the argument exceeds pi/4,
;      . subtract argument from pi
;      . toggle 'cotflg' = 1
;   9. Calculate tangent (see CORDIC description).
;  10. Correct sign of result using 'tnsgn'.
;  11. If 'cotflg' = 1 take reciprocal of result.
;  12. Return F.P. value in w3.
;
;
tan	lda	#0
tan0	sta	cotflg
	sed
	stz	tnsgn
	jsr	tanscale	; reduce argument to range: 0 <= arg <= pi/4
	lda	#<w3	; move scaled argument to w1
	ldy	#>w3
	ldx	#<w1
	jsr	copy2w
	lda	w3+2
	bne	lab049
	rts
lab049	jsr	tanx	; get tangent
	lda	w3
	ora	tnsgn	; correct sign
	sta	w3
	lda	cotflg
	beq	lab050
	lda	#<unit	; take reciprocal if cotangent
	ldy	#>unit
	ldx	#<w1
	jsr	copy2w
	lda	#<w3
	ldy	#>w3
	ldx	#<w2
	jsr	copy2w
	jsr	div
lab050	rts

cot	lda	#1
	jmp	tan0
csc	jsr	sin
	lda	#<w3
	ldy	#>w3
	ldx	#<w2
	jsr	copy2w
	lda	#<unit
	ldy	#>unit
	ldx	#<w1
	jsr	copy2w
	jmp	div

tanx	lda	w1+2	; main tangent routine
	bne	tn0
tn0x	lda	#0
	ldx	#7
lab051	sta	w3,x	; argument is zero, clear 'w3' and return.
	dex
	bpl	lab051
	rts
tn0	bit	w1
	bvc	tn0x1
	lda	w1	; exponent is negative, check size
	and	#$0f
	bne	tn0x1a	; large exponent -- return argument
	lda	w1+1
	cmp	#8
	bcc	tn0x1
tn0x1a	ldx	#7
lab052	lda	w1,x	; large neg exponent -- return argument
	sta	w3,x
	dex
	bpl	lab052
	rts
tn0x1	ldx	#3*wreglen-1
	lda	#0	; clear working registers
tn0xb	sta	ra,x   
	dex
	bpl	tn0xb
	lda	#<kxatn	; initialize arctangent table pointer
	sta	ptr1
	lda	#>kxatn
	sta	ptr1+1

	lda	w1+1
	lsr
	cld
	adc	#5
	sed
	tax
	ldy	#5
tn00b	lda	w1+2,y	; move angle to 'ra' with decimal point alignment  
	sta	ra+2,x
	dex
	dey
	bpl	tn00b
	lda	w1+1
	and	#1
	beq	tn00a
	ldx	#12
	lda	#0
tn00c	ldy	ra+2,x
	ora	tlh,y
	sta	ra+2,x
	lda	thl,y
	dex
	bpl	tn00c
tn00a	jsr	tnpdiv	; call tangent pseudo-divide routine
;
;   're' contains pseudo-quotient in extended format.
;       (00 0D 0D 0D 0D 0D ...)
;
;   Note that the first digit is zero because the argument
;   scaling reduces the argument below pi/4 < 1.0.
;   now call routine to create fixed point tangent
;
	jsr	tnpmul
    
;   On return, 'ra' contains a quantity proportional to 'y', and 'rb'
;   contains a quantity proportional to 'x'. First, convert the quantities
;   from the fixed point form to floating point in 'w1' and 'w3'.
;

	stz	w1	; clear exponent and sign field of 'w1'
	stz	w1+1
	lda	#8      
	sta	bctr	; limit iterations
t0a1	lda	ra+2
	and	#$f0	; is the decimal point OK?
	bne	t0b
	lda	#$40
	sta	w1	; exponent is negative
	inc	w1+1
	lda	#0
	ldx	#12
t0a2	ldy	ra+2,x	; shift 'ra' left and adjust exponent in 'w1'
	ora	tlh,y
	sta	ra+2,x
	lda	thl,y
	dex
	bpl	t0a2
	dec	bctr
	bne	t0a1
	lda	#0
	ldx	#7
t0a3	sta	w3,x	; clear 'w3' to zero
	dex
	bpl	t0a3
	rts
t0b	ldx	#5
t0b1	lda	ra+2,x	; move mantissa to 'w1'
	sta	w1+2,x
	dex
	bpl	t0b1
;
;   check 'rb' position
;
	lda	rb+1
	beq	t0b2
	ldx	#7
t0b1a	lda	w1,x	; 'w1' contains result
	sta	w3,x
	dex
	bpl	t0b1a
	rts
t0b2	lda	#8
	sta	bctr
	lda	#$40
	sta	w2
	stz	w2+1	; 'rb' has negative exponent
t0b2a	lda	rb+2	; shift left until aligned
	and	#$f0
	bne	t0c
	inc	w2+1
	lda	#0
	ldx	#10
t0b3	ldy	rb+2,x
	ora	tlh,y
	sta	rb+2,x
	lda	thl,y
	dex
	bpl	t0b3
	dec	bctr
	bne	t0b2a
	ldx	#7
t0b4	lda	pio2,x	; angle is pi/2  ('rb' is zero)
	sta	w3,x
	dex
	bpl	t0b4
	rts
t0c	ldx	#5
t0c1	lda	rb+2,x	; move mantissa to 'w2'
	sta	w2+2,x
	dex
	bpl	t0c1
	jmp	div	; jsr, rts


;
; tangent pseudo-multiply routine
; enter with 'ra' = residual = y
; set 'rb' = 1.0             = x
; copy 'ra' and 'rb' to 'rc' and 'rd'
;
tnpmul	ldx	#$10
	stx	rb+2
	ldx	#10
lab053	lda	ra+2,x
	sta	rc+2,x
	lda	rb+2,x
	sta	rd+2,x
	dex
	bpl	lab053
	lda	#0
	sta	ctr3	; pseudo-digit index
;
; repeat algorithm for each digit of pseudo-quotient
;
;   ra = y2
;   rb = x2
;   rc = y1
;   rd = x1
;
tnpm1	ldx	ctr3
	dec	re,x
	bmi	tnpm2
	sed
;
;   copy 'ra', 'rb' to 'rc','rd'
;
	ldx	#10
lab054	lda	ra+2,x
	sta	rc+2,x
	lda	rb+2,x
	sta	rd+2,x
	dex
	bpl	lab054

;   now shift 'rc' and 'rd' right 
;
	lda	ctr3
	jsr	scdnrt
;
;   now perform pseudo-multiplication
;
	ldx	#10
	sec
lab055	lda	rb+2,x	; x2 = x1 - y1*10^j
	sbc	rc+2,x
	sta	rb+2,x
	dex
	bpl	lab055
	ldx	#10
	clc
lab056	lda	ra+2,x	; y2 = y1 + x1*10^j
	adc	rd+2,x
	sta	ra+2,x
	dex
	bpl	lab056
	bra	tnpm1
tnpm2	inc	ctr3
	lda	#5
	cmp	ctr3
	bcs	tnpm1
tnpmx	rts


;
;   tangent pseudo-divide routine: calculates CORDIC pseudo-quotient
;   for tangent function.
;
tnpdiv	lda	#5	; number of table entries in arctangent table
	sta	ctr3
	ldx	#0	; index into result register (rd)
	stx	re
tpd0	ldy	#7	; (subtract until underflow, then restore)
	sec
tpd1	lda	ra+2,y	; subtract 1 instance of current table value
	sbc	(ptr1),y
	sta	ra+2,y
	dey
	bpl	tpd1
	bcc	tpd2a	; underflow? no, do another loop
	inc	re,x
	bra	tpd0
tpd2a	ldy	#7	; restore
	clc
tpd2	lda	ra+2,y
	adc	(ptr1),y
	sta	ra+2,y
	dey
	bpl	tpd2
	inx
	lda	#0
	sta	re,x
	cld		; ...bump tangent table pointer to next entry
	clc
	lda	ptr1
	adc	#8
	sta	ptr1
	bcc	tpd3
	inc	ptr1+1
tpd3	sed            
	dec	ctr3
	bpl	tpd0	; loop until each digit of pseudo-quotient is done
	rts


;
;   arccos -- calculates arcosine(x): acos(x) = pi/2 - asin(x)
;
acos	jsr	asin
	lda	#<w3
	ldy	#>w3
	ldx	#<w2
	jsr	copy2w
	lda	#<pio2
	ldy	#>pio2
	ldx	#<w1
	jsr	copy2w
	jmp	sub	; jsr, rts


;                                  x
;   arcsine --  asin(x) = atan( ---------)
;                               û 1 - x^2
asin	lda	w1+2
	bne	lab057
	jmp	retzr
lab057	lda	w1
	and	#$80	; save argument sign (append to result)
	sta	sinflg
	lda	#<w1	; make argument positive
	and	#$7f
	sta	w1
	lda	#<w1	; save to temporary register
	sta	ptr1
	lda	#>w1
	sta	ptr1+1
	lda	#<reg1
	sta	ptr2
	lda	#>reg1
	sta	ptr2+1
	jsr	copyreg
	lda	#<w1        
	ldy	#>w1
	ldx	#<w2
	jsr	copy2w
	jsr	mul	; form arg^2
	lda	#<unit
	ldy	#>unit
	ldx	#<w1
	jsr	copy2w
	lda	#<w3
	ldy	#>w3
	ldx	#<w2
	jsr	copy2w
	jsr	sub         ; form 1 - arg^2
;
;   test result of 1 - arg^2
;
;   if zero, argument is +-1.0 and angle is +-pi/2 (sin) or -pi (cos)
;   if negative, argument is invalid.
;
	lda	w3+2
	beq	argzr
	lda	w3
	bpl	argpls
	jmp	rangerr	; argument > 1.0, return with error
argzr	lda	#<pio2
	ldy	#>pio2
	ldx	#<w3
	jsr	copy2w
	bra	fxsgn
argpls	lda	#<w3
	ldy	#>w3
	ldx	#<w1
	jsr	copy2w
	jsr	sqrt	; form û 1 - arg^2
	lda	#<reg1
	ldy	#>reg1
	ldx	#<w1
	jsr	copy2w
	lda	#<w3
	ldy	#>w3
	ldx	#<w2
	jsr	copy2w
	jsr	div	; form arg / û 1 - arg^2 
	lda	#<w3
	ldy	#>w3
	ldx	#<w1
	jsr	copy2w
	jsr	atan

fxsgn	lda	w3
	ora	sinflg
	sta	w3
	rts


;
;   arctangent 
;   27/04/2015 - M. Ferrer : added case "argument null" (returns 0)
;
;   0. If argument is null, put 0 in w3 and exit.
;   1. Test argument sign,
;       . save sign, make argument positive.
;   2. Test size of argument,
;       . if argument < 1e-6 correct sign and return
;         argument in w3.
;       . if argument > 1.0, set cotflg replace argument
;         with its reciprocal.
;       . else clear cotflg.
;   3. Compute arctangent of argument. (0 <= angle <= pi/4).
;   4. If cotflg, subtract from pi/2.
;   5. Correct sign.
;   6. Return result in w3.
;
atan	lda	w1+2	; is argument null ?
	bne	atan_0

	jmp	retzr	; yes : return zero

atan_0	sed		; no : compute arc tangent
	lda	#0
	sta	cotflg
	sta	tnsgn
	bit	w1
	bpl	lab058	; is argument negative ?
	lda	#$80	; yes, set 'tnsgn' flag
	sta	tnsgn	; and make arg +
	lda	#$4f
	and	w1
	sta	w1
lab058	bvs	argchk
	jsr	inv
	lda	#<w3
	ldy	#>w3
	ldx	#<w1
	jsr	copy2w
	inc	cotflg	; argument is inverted and cotflg is set
argchk	lda	w1	; here, 0 <= arg <= 1.0
	and	#$0f
	bne	toosml
	lda	w1+1
	cmp	#7
	bcc	argok
toosml	lda	#0	; argument is too small to need computation
	cmp	cotflg
	beq	tsxt
	lda	#<pio2
	ldy	#>pio2
	ldx	#<w2
	jsr	copy2w
	jsr	div
tsxt	lda	tnsgn
	ora	w1
	sta	w1
	lda	#<w1
	ldy	#>w1
	ldx	#<w3
	jmp	copy2w	; jsr, rts

argok	ldx	#5*wreglen-1
	lda	#0
atn00a	sta	ra,x	; clear working storage: ra, rb, rc, rd and re
	dex
	bpl	atn00a
;
;   move argument to ra with decimal point alignment
;
	lda	w1+1
	lsr
	cld
	clc
	adc	#5
	sed
	tax
	ldy	#5
lab059	lda	w1+2,y
	sta	ra+2,x
	dex
	dey
	bpl	lab059
	lda	w1+1
	and	#1
	beq	lab060
	ldx	#7	; even exponent, shift right one digit
	lda	#0
atn1	ldy	ra+2,x
	ora	tlh,y
	sta	ra+3,x
	lda	thl,y
	dex
	bpl	atn1
	sta	ra+2
lab060	ldx	#0
	stx	ctr3	; storage for pseudo-quotient index
	lda	#$10	; 'ra' and 'rb' are now initialized 
	sta	rb+2
;
;   ra = y2 = arg (0 < arg < pi/4)
;   rb = x2 = 1.0
;
;   1. initialize pseudo-quotient digit counter and index
;       ctr3 <- index = 0, re used for pq
;   2. outer loop
;      (A) inner loop
;        y2 -> y1   (rc)
;        x2 -> x1   (rd)
;        y2 shr index (10^j)  (1st pass, j=0, no shift required)
;        x2 shr index (10^j)    "   "     "        "      "
;        y2 = y2 - x1
;        if y2 < 0 goto (B)
;        x2 = x2 + y1
;        re[index]++
;        goto (A)
;     (B) y2 = y2 + x1  (restore after subtraction underflow)
;        index++
;        if index is LT 7 goto 2.
;

	stz	ctr3
atnlp1	ldx	#6
lab061	lda	ra+2,x	; y2,x2 -> y1,x1
	sta	rc+2,x
	lda	rb+2,x
	sta	rd+2,x
	dex
	bpl	lab061
	lda	ctr3
	beq	lab062
	jsr	scdnrt	; shift y1,x1 to correct postion
lab062	sed
	sec
	ldx	#6
lab063	lda	ra+2,x
	sbc	rd+2,x
	sta	ra+2,x
	dex
	bpl	lab063
	bcc	rstr	; underflow: do restore
	clc
	ldx	#6
lab064	lda	rb+2,x
	adc	rc+2,x
	sta	rb+2,x
	dex
	bpl	lab064
	ldx	ctr3
	inc	re,x	; update current pseudo-quotient digit
	bra	atnlp1
rstr	ldx	#6
	clc
lab065	lda	ra+2,x
	adc	rd+2,x
	sta	ra+2,x
	dex
	bpl	lab065
	inc	ctr3
	lda	#6	; number of desired pseudo-quotient digits
	cmp	ctr3
	bcs	atnlp1
;
;   pseudo-quotient digits are now in 're' (6 digits)
;   1. Convert 'ra' to F.P. in w1
;   2. Convert 'rb' to F.P. in w2
;   3. jsr divide, result in w3 is residual starting value for next step
;   4. move w3 to 'ra' with decimal point alignment
;   5. pseudo-multiply. set 'i' to 0.
;       A. for each digit[i] in pseudo-quotient, add kxatnx[i]
;       B. decrement digit
;       C. digit = 0?
;           no? goto A.
;           yes? increment 'i'
;       D. convert 'ra' to F.P. in w3, return
;
	lda	#10
	sta	bctr	; guard counter
	lda	#$40
	sta	w1
	stz	w1+1	; 'rs' has negative exponent
lab066	lda	ra+2	; shift left until aligned
	and	#$f0
	bne	atnres
	inc	w1+1
	lda	#0
	ldx	#10
lab067	ldy	ra+2,x
	ora	tlh,y
	sta	ra+2,x
	lda	thl,y
	dex
	bpl	lab067
	dec	bctr
	bne	lab066
atnres	ldx	#5
lab068	lda	ra+2,x
	sta	w1+2,x
	dex
	bpl	lab068
	ldx	#7
lab069	lda	rb,x
	sta	w2,x
	dex
	bpl	lab069
	jsr	div
;
;   pseudo-multiply for arctangent
;
;   1. copy w3 to 'ra' with decimal point aligned
;   2. initialize kxatn table pointer in 'ptr1'
;   3. initialize pseudo-quotient index in 'ctr3'
;   4. pseudo-multiply by repeated addition:
;       A. for count given by each digit of pseudo-quotient,
;           add current table entry to running sum in 'ra',
;       B. increment index (ctr3) and bump table pointer
;           to next entry.
;       C. if index < 7, goto A.
;
	ldx	#15
	lda	#0
lab070	sta	ra,x	; clear 'ra'
	dex
	bpl	lab070
	lda	w3
	and	#$0f
	bne	atnpm
	lda	w3+1
	cmp	#9
	bcs	atnpm
	lsr
	clc
	cld
	adc	#5
	tax
	ldy	#5
lab071	lda	w3+2,y
	sta	ra+2,x
	dex
	dey
	bpl	lab071
	lda	#<w1+1
	and	#1
	beq	atnpm
	ldx	#7	; shift right one digit
	lda	#0
lab072	ldy	ra+2,x
	ora	tlh,y
	sta	ra+3,x
	lda	thl,y
	dex
	bpl	lab072
	sta	ra+2
atnpm	lda	#<kxatn	; setup table point
	sta	ptr1
	lda	#>kxatn
	sta	ptr1+1
	ldx	#0	; initialize pseudo-quotient index
	stx	ctr3
apmlp	dec	re,x
	bmi	atnxtd
	ldy	#7
	sed
	clc
lab073	lda	ra+2,y
	adc	(ptr1),y
	sta	ra+2,y
	dey
	bpl	lab073
	bmi	apmlp
atnxtd	inc	ctr3
	cld
	clc
	lda	ptr1
	adc	#8
	sta	ptr1
	bcc	lab074
	inc	ptr1+1
lab074	ldx	ctr3
	cpx	#7	; number of pseudo-quotient digits
	bcc	apmlp
	stz	w3
	stz	w3+1	; clear result exponent field
atn02a	lda	ra+2
	and	#$f0
	bne	atn03
	lda	#$40	; negative exponent
	sta	w3
	ldx	#15
	lda	#0
atn02b	ldy	ra+2,x	;shift 'ra' left until normalized
	ora	tlh,y
	sta	ra+2,x
	lda	thl,y
	dex
	bpl	atn02b
	inc	w3+1
	bra	atn02a
atn03	ldx	#5
lab075	lda	ra+2,x
	sta	w3+2,x
	dex
	bpl	lab075
	lda	cotflg
	beq	atnx
	lda	#<pio2
	ldy	#>pio2
	ldx	#<w1
	jsr	copy2w
	lda	#<w3
	ldy	#>w3
	ldx	#<w2
	jsr	copy2w
	jsr	sub
atnx	lda	w3
	ora	tnsgn
	sta	w3
	rts


;
;   arc cosecant -- acsc(x) = asin(1/x)
;
;   w3 = acsc(w1)
;
acsc	lda	#<w1
	ldy	#>w1
	ldx	#<w2
	jsr	copy2w
	lda	#<unit
	ldy	#>unit
	ldx	#<w1
	jsr	copy2w
	jsr	div
	lda	#<w3		; M. Ferrer : missing code
	ldy	#>w3		; we need to put previous div
	ldx	#<w1		; result back to w1
	jsr	copy2w
	jmp	asin		; jsr, rts


;
;   arc secant -- asec(x) = acos(1/x)
;
;   w3 = asec(w1)
;
asec	lda	#<w1
	ldy	#>w1
	ldx	#<w2
	jsr	copy2w
	lda	#<unit
	ldy	#>unit
	ldx	#<w1
	jsr	copy2w
	jsr	div
	lda	#<w3		; M. Ferrer : missing code
	ldy	#>w3		; we need to put previous div
	ldx	#<w1		; result back to w1
	jsr	copy2w
	jmp	acos		; jsr, rts


;
;   arc cotangent -- acot(x) = atan(1/x)
;
;   w3 = acot(w1)
;
acot	lda	#<w1
	ldy	#>w1
	ldx	#<w2
	jsr	copy2w
	lda	#<unit
	ldy	#>unit
	ldx	#<w1
	jsr	copy2w
	jsr	div
	lda	#<w3		; M. Ferrer : missing code
	ldy	#>w3		; we need to put previous div
	ldx	#<w1		; result back to w1
	jsr	copy2w
	jmp	atan		; jsr, rts


;
;   loge
;
;   calculate the natural logarithm (ln) of the argument in 'w1'.
;
;   w3 = ln(w1)
;   
loge	lda	w1
	bpl	lab076	; test for negative number
	jmp	domnerr
lab076	lda	w1+2
	bne	lab077	; test for zero argument
	jmp	rangerr
lab077	ldx	#7
lab078	lda	w1,x	; save argument
	sta	reg1,x
	dex
	bpl	lab078
	lda	w1	; if argument is near unity, use series approximation
	bne	lab079	;  9.9 e-1 < arg < 1.01 e0
	lda	w1+1
	bne	lab079
	lda	w1+2
	cmp	#$10
	bne	lab079
;
;   argument is near unity. test for series solution
;
	lda	#<unit
	ldy	#>unit
	ldx	#<w2
	jsr	copy2w
	jsr	sub
	lda	w3+2
	bne	lab079	; check for arg = 0.0
	jmp	retzr
lab079	lda	w3
	cmp	#$40
	bne	lnargrst
	lda	w3+1
	cmp	#$3
	bcc	lnargrst
	jmp	lnser
lnargrst	lda	#<reg1
	ldy	#>reg1
	ldx	#<w1
	jsr	copy2w
lab080	jsr	rr_clr	; clear all workspace regs
	ldx	#5	; copy mantissa to 'ra'
lab081	lda	w1+2,x
	sta	ra+2,x
	dex
	bpl	lab081
	stz	ctr1	; 'j' index for pseudo-divide
	lda	#6
	sta	ctr3
;
;   main loop
;
lnlp0	jsr	cpyra2rc	; compute each pseudo-quotient digit
	lda	ctr1
	jsr	shfrcn
	jsr	addrc2ra
	bcs	lab082
	ldx	ctr1
	inc	rd,x
	bra	lnlp0

lab082	jsr	subrc2ra	; restore remainder after underflow
	inc	ctr1           
	dec	ctr3
	bne	lnlp0	; another iteration?

	ldx	#6
	sec
lab083	lda	#0	; compute residual to start pseudo-multiply
	sbc	ra+2,x
	sta	ra+2,x
	dex
	bpl	lab083

	ldx	#7	; shift right one digit
	lda	#0
lab084	ldy	ra+2,x
	ora	tlh,y
	sta	ra+3,x
	lda	thl,y
	dex
	bpl	lab084
	sta	ra+2
;
;   pseudo-multiply
;
	lda	#<kxlog	; set pointer to table of logs
	sta	ptr1
	lda	#>kxlog
	sta	ptr1+1
	stz	ctr1	; initialize counters
	lda	#6
	sta	ctr2
lnlp1	ldx	ctr1	; perform pseudo-multiply for each table entry
	dec	rd,x
	bmi	npsm
	ldy	#6
	clc
lab085	lda	ra+2,y
	adc	(ptr1),y
	sta	ra+2,y
	dey
	bpl	lab085
	bmi	lnlp1
npsm	cld	; bump table pointer to next coefficient
	clc
	lda	ptr1
	adc	#8
	sta	ptr1
	bcc	lab086
	inc	ptr1+1
lab086	sed
	inc	ctr1
	dec	ctr2
	bne	lnlp1
;
;   'ra' contains ln(mantissa). adjust exponent if needed
;
lab087	lda	ra+2	; (ctr2 is now 0)
	and	#$f0
	bne	ln2fp
	jsr	shflfta
	inc	ctr2
	bne	lab087
ln2fp	ldx	#5	; save current logarithm as F.P. in reg2
lab088	lda	ra+2,x
	sta	reg2+2,x
	dex
	bpl	lab088
	lda	#$80	; sign of mantissa?
	sta	reg2
	lda	ctr2
	sta	reg2+1
	beq	lab089
	lda	#$40	; sign of exponent
	ora	reg2
	sta	reg2
lab089	jsr	exp2fp
	lda	#<unit
	ldy	#>unit
	ldx	#<w2
	jsr	copy2w
	jsr	add
	lda	#<w3
	ldy	#>w3
	ldx	#<w1
	jsr	copy2w
	lda	#<ln10
	ldy	#>ln10
	ldx	#<w2
	jsr	copy2w
	jsr	mul
	ldx	#7
lab090	lda	reg2,x
	sta	w1,x
	dex
	bpl	lab090
	lda	#<w3
	ldy	#>w3
	ldx	#<w2
	jsr	copy2w
	jmp	add	; jsr, rts

;
;   exp2fp -- convert exponent in 'w1' to signed F.P. number in 'w1'
;
exp2fp	ldx	#5
	lda	#0
lab091	sta	w1+2,x	; clear mantissa space
	dex
	bpl	lab091
	lda	w1
	and	#$7f
	sta	w1
	and	#$40
	asl
	tax		; save sign bit
	lda	w1
	and	#$f
	bne	bigexp
	stx	w1
	lda	w1+1
	and	#$f0
	bne	medexp
	lda	w1+1
	asl
	asl
	asl
	asl
	sta	w1+2
	lda	#0
	sta	w1+1
	stx	w1
	rts

medexp	sta	w1+2
	lda	#1
	sta	w1+1
	stx	 w1
	rts

bigexp	asl
	asl
	asl
	asl
	sta	w1+2
	lda	w1+1
	lsr
	lsr
	lsr
	lsr
	ora	w1+2
	sta	w1+2
	lda	w1+1
	asl
	asl
	asl
	asl
	sta	w1+3
	lda	#2
	sta	w1+1
	stx	w1
	rts

cpyra2rc	ldy	#6
lab092	lda	ra+2,y
	sta	rc+2,y
	dey
	bpl	lab092
	rts
;
;   add 'rc' to 'ra'. 'rc' is shifted right by ctr1 digits
;
addrc2ra	cld
	clc
	ldx	#7
	sed
lab093	lda	ra+2,x
	adc	rc+2,x
	sta	ra+2,x
	dex
	bpl	lab093
	rts
;
;   subtract 'rc' from 'ra'. 'rc' is shifted right by ctr1 digits
;
subrc2ra	cld
	sec
	sed
	ldx	#7
lab094	lda	ra+2,x
	sbc	rc+2,x
	sta	ra+2,x
	dex
	bpl	lab094
	rts
;
;   shift 'ra' left one digit
;
shflfta	ldx	#8
	lda	#0
lab095	ldy	ra+2,x
	ora	tlh,y
	sta	ra+2,x
	lda	thl,y
	dex
	bpl	lab095
	rts

shfrcn	sta	ctr2	; shift rc right n digits ('n' in a_reg)        
	and	#1	; is it an odd number?
	beq	src0	; no, only do byte shifts
	lda	#0	; yes, do single digit shift
	ldx	#7
lab096	ldy	rc+2,x
	ora	tlh,y
	sta	rc+3,x
	lda	thl,y
	dex
	bpl	lab096
	sta	rc+2
src0	lda	ctr2
	lsr		; convert to bytes
	beq	srcx	; exit if no further shift required
	cld
	clc
	adc	#7
	sed
	tay
	ldx	#7
lab097	lda	rc+2,x
	sta	rc+2,y
	dey
	dex
	bpl	lab097
	lda	#0
lab098	sta	rc+2,y
	dey
	bpl	lab098
srcx	rts
;                                          
;   lnser -- return logarithm of number near unity by series expansion
;            enter with (arg - 1) in 'w3'.
;
lnser	lda	#<kln
	sta	ptr3
	lda	#>kln
	sta	ptr3+1
	lda	#3
	sta	ctr1
	ldx	#7
lab099	lda	w3,x	; save argument (x) to 'reg2'
	sta	reg2,x
	dex
	bpl	lab099
	ldy	#7
lab100	lda	(ptr3),y	; preload 'w3' with 1st coefficient
	sta	w3,y
	dey
	bpl	lab100
; main loop
lnslp	ldy	#7
lab101	lda	w3,y
	sta	w1,y
	lda	reg2,y
	sta	w2,y
	dey
	bpl	lab101
	jsr	mul	; multiply  
	dec	ctr1
	bmi	lnsrx
	cld
	clc
	lda	ptr3
	adc	#8
	sta	ptr3
	bcc	lab102
	inc	ptr3+1
lab102	sed
	ldy	#7
lab103	lda	(ptr3),y
	sta	w1,y
	lda	w3,y
	sta	w2,y
	dey
	bpl	lab103
	jsr	add	; add next coefficient
	bra	lnslp
lnsrx	rts


;
;   exp
;
;   calculate exp(arg) where the argument is in 'w1'.
;
;   First, the argument is checked for sign. If the sign of the
;   argument is negative, the absolute value of the argument
;   is sent to the 'exp' routine and the reciprocal of the result
;   is returned to the caller.
;
;   Next, the integer part of the argument is taken and used to
;   calculate a power-of-ten multiplier (number of decimal places)
;   is found. A residual decimal fraction from this operation is
;   added to the fractional value of the given argument and the
;   'exp' routine is called with this sum as its argument.
;
;   The result is combined with the previously calculated decimal
;   point adjustment and a test for taking the reciprocal is made.
;   After all operations are complete, the result is returned to
;   the caller.
;
;   This implementation takes the mantissa and treats it as if
;   it were in the range: 0.1 <= m <= 0.99999.... instead of
;   1.0 <= m <= 9.99999......
;
;   The algorithm works by finding a pseudo-quotient from the
;   mantissa using a table of natural logs of 2, 1.1, 1.01, 1.001,
;   1.0001, etc. The digits of the pseudo-quotient represent the
;   number of subtractions of the appropriate log before underflow.
;
;   After 6-8 pseudo-quotient digits have been found, 1.0 is added
;   to the residual and the pseudo-multiply algorithm is run.
;
;   q = arg/ln10
;   f = floor(q)
;   r = q - f
; 
;
exp	stz	expflg
	lda	w1
	and	#$0f
	beq	lab104
xpre	jmp	rangerr
lab104	lda	w1+2
	bne	lab105
xrt1	lda	#<unit
	ldy	#>unit
	ldx	#<w3
	jmp	copy2w
lab105	bit	w1
	bpl	lab106
	lda	#$80
	sta	expflg
	eor	w1
	sta	w1
lab106	bit	w1
	bvc	lab107
	lda	w1	; check for range problem with exponent
	and	#$0f
	bne	xrt1	; large negative exponent returns 1.0000...
	lda	w1+1
	cmp	#$12
	bcs	xrt1
	bcc	lab108
lab107	lda	w1+1  
	cmp	#4
	bcs	xpre
	cmp	#3
	bne	lab108
	lda	w1+2
	cmp	#$23
	bcs	xpre
lab108	lda	#<iln10	; form arg/ln10
	ldy	#>iln10
	ldx	#<w2
	jsr	copy2w
	jsr	mul
	lda	#<w3
	ldy	#>w3
	ldx	#<reg1
	jsr	copy2w	; save backup copy of arg/ln10 in 'reg1'
	lda	#<w3
	ldy	#>w3
	ldx	#<w1
	jsr	copy2w
	jsr	int
	lda	#<w3
	ldy	#>w3
	ldx	#<w2
	jsr	copy2w
	lda	#<w3
	ldy	#>w3
	ldx	#<reg2	; save power of 10 (exponent) in 'reg2'
	jsr	copy2w
	jsr	sub
	lda	#<w3
	ldy	#>w3
	ldx	#<w1
	jsr	copy2w
	lda	#<ln10
	ldy	#>ln10
	ldx	#<w2  
	jsr	copy2w
	jsr	mul
	lda	#0
	sta	tmp1	; clear shift count register
	bit	w3
	bvc	xshft
	lda	w3+1
	cmp	#$12
	bcc	lab109
	jmp	xrt1
lab109	tax
	lda	dec2bin,x
	sta	tmp1
xshft	jsr	xexp
;
;   mantissa of exponential is in 'w1'. get exponent from 'reg2'
;
	ldy	reg2+1
	lda	#0
	sta	reg2+1
xxp	ldx	#3	; move exponent value from mantissa to exponent
lab110	asl	reg2+3
	rol	reg2+2
	rol	reg2+1
	rol	reg2
	dex
	bpl	lab110
	dey
	bpl	xxp
	lda	reg2	; copy exponent to w1
	sta	w3
	lda	reg2+1
	sta	w3+1
	bit	expflg
	bpl	lab111
	lda	#<w3
	ldy	#>w3
	ldx	#<w2
	jsr	copy2w
	lda	#<unit
	ldy	#>unit
	ldx	#<w1
	jsr	copy2w
	jsr	div
lab111	rts
;
;   xexp -- main exponential computation routine
;
xexp	jsr	rr_clr
	ldx	#5
	lda	tmp1
	lsr	; ignore odd bit for now
	cld
	clc
	adc	#5
	tay
lab112	lda	w3+2,x
	sta	ra+2,y
	sta	rb+2,x
	dey
	dex
	bpl	lab112
	lda	tmp1
	sta	re
	lsr		; do odd bit shift (if any)
	bcc	lab114

	ldx	#8
	lda	#0
lab113	ldy	ra+1,x	;   12 34 00
	ora	tlh,y	;   01 23 40
	sta	ra+2,x
	lda	thl,y
	dex
	bpl	lab113
	sta	ra+2,x
lab114
;
;   this is the routine which computes the natural logarithm of
;   the number in 'ra'.
;
xnl	lda	#<kxlog
	ldy	#>kxlog
	sta	ptr1
	sty	ptr1+1
	lda	#7
	sta	tmp2	; 'tmp2' is outer loop counter
	stz	ctr1
;
;   subtract 'kxlog' entry from 'ra' until underflow
;
mxlp0	lda	#0
	tax
mxlp	sed
	sec
	ldy	#7
lab115	lda	ra+2,y
	sbc	(ptr1),y
	sta	ra+2,y
	sta	rc+2,y
	dey
	bpl	lab115
	inx
	bcs	mxlp
	dex
	txa		; 'x' has current pseudo-quotient digit
	ldx	ctr1
	sta	rd,x
	inc	ctr1
	clc
	ldy	#7
lab116	lda	ra+2,y	; restore 'ra' after underflow
	adc	(ptr1),y
	sta	ra+2,y
	dey
	bpl	lab116
	cld
	clc
	lda	ptr1
	adc	#8
	sta	ptr1
	bcc	lab117
	inc	ptr1+1
lab117	dec	tmp2
	bpl	mxlp0
;
;   pseudo-quotients are found, now perform pseudo-multiply
;
	stz	ctr1	; 'ctr1' is offset to pseudo-quotient digit
	lda	#$10	; add '1' to residual in 'ra'    
	sta	ra+2
	ldx	#7
	stx	ctr3	; 'ctr3' is outer loop counter (number of digits)
	ldx	#7
lab118	lda	ra+2,x	; copy 'ra' to 'rc'
	sta	rc+2,x
	dex
	bpl	lab118
;
;  multiply 'ra' by 2.0, 1.1, 1.01, etc. iterate each multiplier
;   as controlled by pseudo-digit in 'rd'
;
xml	ldx	ctr1
	dec	rd,x
	bmi	xbd
;
;   copy 'ra' to 'rc' shifted right
;
	ldx	#7
lab119	lda	ra,x
	sta	rc,x
	dex
	bpl	lab119
	lda	ctr1
	jsr	shfrcn
	sed
	ldx	#5
	clc
lab120	lda	ra+2,x 
	adc	rc+2,x
	sta	ra+2,x
	dex
	bpl	lab120
	bra	xml
xbd	inc	ctr1
	dec	ctr3
	bpl	xml
	ldx	#5
lab121	lda	ra+2,x
	sta	w3+2,x
	dex
	bpl	lab121
	lda	#0
	sta	w3
	sta	w3+1
	rts


;
;   hyberbolic sine -- calculates sinh(x) = (exp(x)-exp(-x))/2
;
;   w3 = sinh(w1)
;
sinh	jsr	exp
	lda	#<w3
	ldy	#>w3
	ldx	#<w2
	jsr	copy2w
	lda	#<unit
	ldy	#>unit
	ldx	#<w1
	jsr	copy2w
	jsr	div
	lda	#<w3
	ldy	#>w3
	ldx	#<w1
	jsr	copy2w
	lda	w1
	ora	#$80
	sta	w1
	jsr	add
	lda	#<w3
	ldy	#>w3
	ldx	#<w1
	jsr	copy2w
	lda	#<half
	ldy	#>half
	ldx	#<w2
	jsr	copy2w
	jmp	mul	; jsr, rts


;
;   hyperbolic cosine -- calculates cosh(x) = (exp(x) + exp(-x))/2
;
;   w3 = cosh(w1)
;
cosh	jsr	exp
	lda	#<w3
	ldy	#>w3
	ldx	#<w2
	jsr	copy2w
	lda	#<unit
	ldy	#>unit
	ldx	#<w1
	jsr	copy2w
	jsr	div
	lda	#<w3
	ldy	#>w3
	ldx	#<w1
	jsr	copy2w
	jsr	add
	lda	#<w3
	ldy	#>w3
	ldx	#<w1
	jsr	copy2w
	lda	#<half
	ldy	#>half
	ldx	#<w2
	jsr	copy2w
	jmp	mul	; jsr, rts


;
;   tanh -- compute (exp(x) - exp(-x))/(exp(x) + exp(-x))
;
;   w3 = tanh(w1)
;
tanh	jsr	exp
	lda	#<w3	; w3 = exp(x)
	ldy	#>w3
	ldx	#<w2
	jsr	copy2w	; w2 = exp(x)
	ldx	#7
lab122	lda	w3,x	; save exp(x) in reg2
	sta	reg2,x
	dex
	bpl	lab122

	lda	#<unit
	ldy	#>unit
	ldx	#<w1
	jsr	copy2w
	jsr	div	; w3 = exp(-x)
	lda	#<w3
	ldy	#>w3
	ldx	#<w2
	jsr	copy2w	; w2 = exp(-x)
	ldx	#7
lab123	lda	w3,x	; save exp(-x) in reg3
	sta	reg3,x
	dex
	bpl	lab123
	lda	#<reg2
	ldy	#>reg2
	ldx	#<w1
	jsr	copy2w
	jsr	sub	; w3 = exp(x) - exp(-x)
	ldx	#7
lab124	lda	reg2,x	; w1 = exp(x)
	sta	w1,x
	dex
	bpl	lab124
	ldx	#7
lab125	lda	reg3,x	; w2 = exp(-x)
	sta	w2,x
	dex
	bpl	lab125
	ldx	#7
lab126	lda	w3,x	; reg2 = exp(x) - exp(-x)
	sta	reg2,x
	dex
	bpl	lab126
	jsr	add	; w3 = exp(x) + exp(-x)
	ldx	#7
lab127	lda	w3,x	; w1 = exp(x) + exp(-x)
	sta	w2,x
	dex
	bpl	lab127
	ldx	#7
lab128	lda	reg2,x	; w2 = exp(x) - exp(-x)
	sta	w1,x
	dex
	bpl	lab128
	jmp	div	; jsr, rts


;
;   hyperbolic cosecant -- csch(x) = 1/sinh(x)
;
;   w3 = csch(w1)
;
csch	jsr	sinh
	lda	#<w3
	ldy	#>w3
	ldx	#<w2
	jsr	copy2w
	lda	#<unit
	ldy	#>unit
	ldx	#<w1
	jsr	copy2w
	jmp	div	; jsr, rts


;
;   hyperbolic secant -- sech(x) = 1/cosh(x)
;
;   w3 = cosh(w1)
;
sech	jsr	cosh
	lda	#<w3
	ldy	#>w3
	ldx	#<w2
	jsr	copy2w
	lda	#<unit
	ldy	#>unit
	ldx	#<w1
	jsr	copy2w
	jmp	div	; jsr, rts


;
;   hyperbolic cotangent -- coth(x) = 1/tanh(x)
;
;   w3 = coth(w1)
;
coth	jsr	tanh
	lda	#<w3
	ldy	#>w3
	ldx	#<w2
	jsr	copy2w
	lda	#<unit
	ldy	#>unit
	ldx	#<w1
	jsr	copy2w
	jmp	div	; jsr, rts


;
;   hyperbolic arc sine -- asinh(x) = ln(x + sqrt(x^2 + 1))
;
;   w3 = asinh(w1)
;
asinh	ldx	#7
lab129	lda	w1,x	; save argument 'x' in reg1
	sta	reg1,x
	dex
	bpl	lab129
	lda	#<w1
	ldy	#>w1
	ldx	#<w2
	jsr	copy2w
	jsr	mul	; form x^2
	lda	#<w3
	ldy	#>w3
	ldx	#<w2
	jsr	copy2w
	lda	#<unit
	ldy	#>unit
	ldx	#<w1
	jsr	copy2w
	jsr	add	; form 1 + x^2
	lda	#<w3
	ldy	#>w3
	ldx	#<w1
	jsr	copy2w
	jsr	sqrt	; w3 = sqrt(1 + x^2)
	lda	#<w3
	ldy	#>w3
	ldx	#<w2
	jsr	copy2w
	lda	#<reg1
	ldy	#>reg1
	ldx	#<w1
	jsr	copy2w
	jsr	add	; w3 = x + sqrt(1 + x^2)
	lda	#<w3
	ldy	#>w3
	ldx	#<w1
	jsr	copy2w
	jmp	loge	; jsr, rts


;
;   hypberbolic arc cosine: acosh(x) = ln(x + sqrt(x^2 - 1))
;
acosh	ldx	#7
lab130	lda	w1,x	; save argument 'x' in reg1
	sta	reg1,x
	dex
	bpl	lab130
	lda	#<w1
	ldy	#>w1
	ldx	#<w2
	jsr	copy2w
	jsr	mul	; form x^2
	lda	#<w3
	ldy	#>w3
	ldx	#<w1
	jsr	copy2w
	lda	#<unit
	ldy	#>unit
	ldx	#<w2
	jsr	copy2w
	jsr	sub	; form x^2 - 1
	lda	#<w3
	ldy	#>w3
	ldx	#<w1
	jsr	copy2w
	jsr	sqrt	; w3 = sqrt(x^2 - 1)
	lda	#<w3
	ldy	#>w3
	ldx	#<w2
	jsr	copy2w
	lda	#<reg1
	ldy	#>reg1
	ldx	#<w1
	jsr	copy2w
	jsr	add	; w3 = x + sqrt(x^2 - 1)
	lda	#<w3
	ldy	#>w3
	ldx	#<w1
	jsr	copy2w
	jmp	loge	; jsr, rts


;                                           1 + x
;   hyperbolic arc tangent:  atanh(x) = ln( -----)
;                                           1 - x
atanh	ldx	#7
lab131	lda	w1,x	; save 'x' in reg1
	sta	reg1,x
	dex
	bpl	lab131
	lda	#<unit
	ldy	#>unit
	ldx	#<w2
	jsr	copy2w
	jsr	add	; w3 = 1 + x
	lda	#<unit
	ldy	#>unit
	ldx	#<w1
	jsr	copy2w
	lda	#<reg1
	ldy	#>reg1
	ldx	#<w2
	jsr	copy2w
	ldx	#7
lab132	lda	w3,x	; save (1 + x) in reg1
	sta	reg1,x
	dex
	bpl	lab132
	jsr	sub	; w3 = 1 - x
	lda	#<w3
	ldy	#>w3
	ldx	#<w2
	jsr	copy2w
	lda	#<reg1
	ldy	#>reg1
	ldx	#<w1
	jsr	copy2w
	jsr	div	; w3 = (1 + x)/(1 - x)
	lda	#<w3
	ldy	#>w3
	ldx	#<w1
	jsr	copy2w
	jsr	loge
	lda	#<w3
	ldy	#>w3
	ldx	#<w1
	jsr	copy2w
	lda	#<half
	ldy	#>half
	ldx	#<w2
	jsr	copy2w
	jmp	mul	; jsr, rts


;
;   hyperbolic arc cosecant: acsch(x) =  asinh(1/x)
;
acsch	lda	#<w1
	ldy	#>w1
	ldx	#<w2
	jsr	copy2w
	lda	#<unit
	ldy	#>unit
	ldx	#<w1
	jsr	copy2w
	jsr	div
	lda	#<w3
	ldy	#>w3
	ldx	#<w1
	jsr	copy2w
	jmp	asinh	; jsr, rts


;
;   hyperbolic arc secant: asech(x) = acosh(1/x)
;
asech	lda	#<w1
	ldy	#>w1
	ldx	#<w2
	jsr	copy2w
	lda	#<unit
	ldy	#>unit
	ldx	#<w1
	jsr	copy2w
	jsr	div
	lda	#<w3
	ldy	#>w3
	ldx	#<w1
	jsr	copy2w
	jmp	acosh	; jsr, rts


;
;   hyperbolic arc cotangent: acoth(x) = atanh(1/x)
;
acoth	lda	#<w1
	ldy	#>w1
	ldx	#<w2
	jsr	copy2w
	lda	#<unit
	ldy	#>unit
	ldx	#<w1
	jsr	copy2w
	jsr	div
	lda	#<w3
	ldy	#>w3
	ldx	#<w1
	jsr	copy2w
	jmp	atanh	; jsr, rts


;
;   log2(x) = loge(x)/loge(2)
;     
log2	jsr	loge
	lda	#<w3
	ldy	#>w3
	ldx	#<w1
	jsr	copy2w
	lda	#<ln2
	ldy	#>ln2
	ldx	#<w2
	jsr	copy2w
	jmp	div	; jsr, rts


;
;   log10(x) = loge(x)/loge(10)
;
log10	jsr	loge
	lda	#<w3
	ldy	#>w3
	ldx	#<w1
	jsr	copy2w
	lda	#<ln10
	ldy	#>ln10
	ldx	#<w2
	jsr	copy2w
	jmp	div	; jsr, rts


;
;   power(x,y) = x^y = exp(y.loge(x)) = w1^w2
pow	lda	#<w2
	sta	ptr1
	lda	#>w2
	sta	ptr1+1
	lda	#<reg3
	sta	ptr2
	lda	#>reg3
	sta	ptr2+1
	jsr	copyreg
	jsr	loge
	lda	#<w3
	ldy	#>w3
	ldx	#<w1
	jsr	copy2w
	lda	#<reg3
	ldy	#>reg3
	ldx	#<w2
	jsr	copy2w
	jsr	mul
	lda	#<w3
	ldy	#>w3
	ldx	#<w1
	jsr	copy2w
	jmp	exp	; jsr, rts


;
;   inv -- reciprocal of x = 1/x
;
;   w3 = 1/w1
;
inv	lda	#<w1
	ldy	#>w1
	ldx	#<w2
	jsr	copy2w
	lda	#<unit
	ldy	#>unit
	ldx	#<w1
	jsr	copy2w
	jmp	div	; jsr, rts


;
;   abs -- w3 = abs(w1)
;
abs	lda	#<w1
	ldy	#>w1
	ldx	#<w3
	jsr	copy2w
	lda	w3
	and	#$7f
	sta	w3
	rts


;
;   chs -- w3 = -w1
;
chs	lda	#<w1
	ldy	#>w1
	ldx	#<w3
	jsr	copy2w
	lda	w3
	eor	#$80
	sta	w3
	rts

;
;   shift 'rc' and 'rd' right 'n' digits. Enter with 'n' in 'a-reg'
;
scdnrt	sta	ctr2
	and	#1	; is it an odd number?
	beq	scd0	; no, only do byte shifts
	lda	#0	; yes, do single digit shift
	ldx	#7
lab134	ldy	rc+2,x
	ora	tlh,y
	sta	rc+3,x
	lda	thl,y
	dex
	bpl	lab134
	sta	rc+2
	lda	#0
	ldx	#7
lab135	ldy	rd+2,x
	ora	tlh,y
	sta	rd+3,x
	lda	thl,y
	dex
	bpl	lab135
	sta	rd+2
scd0	lda	ctr2
	lsr		; convert to bytes
	beq	scdx	; exit if no further shift required
	cld
	clc
	adc	#7
	sed
	tay
	ldx	#7
lab136	lda	rc+2,x
	sta	rc+2,y
	lda	rd+2,x
	sta	rd+2,y
	dey
	dex
	bpl	lab136
	lda	#0
lab137	sta	rc+2,y
	sta	rd+2,y
	dey
	bpl	lab137
scdx	rts


;
;   int -- returns the integer part of a floating point number
;
;   enter with F.P. number in w1, return with result in w3
;
int	lda	#$0F
	bit	w1	; test for sign of exponent and high order exponent digit
	bvs	int0	; negative exponent means integer part is zero
	bne	int1	; large, positive exponent means number is already integer
	jsr	int1	; copy to result register (w3)
	ldx	w1+1
	cpx	#$12
	bcs	intx
	cld
	lda	#$0B
	sec
	sbc	dec2bin,x	; number of zeros after decimal point
	sed
	lsr
	bne	inty
	ldy	#5
	bra	inta
inty	tax
	ldy	#5
	lda	#0
lab138	sta	w3+2,y
	dey
	dex
	bne	lab138
inta	bcc	intx
	lda	#$f0
	and	w3+2,y
	sta	w3+2,y
intx	rts

int0	jmp	retzr	; jsr, rts

int1	lda	#<w1
	ldy	#>w1
	ldx	#<w3
	jmp	copy2w	; jsr, rts


;
;   frac -- returns with fractional part of floating point number
;           Uses brute-force method by substracting integer part
;
frac	ldx	#7
lab139	lda	w1,x	; save 'x' in reg3
	sta	reg3,x
	dex
	bpl	lab139
	jsr	int	; find integer part of argument in w1
	lda	#<w3	; move integer part to w2
	ldy	#>w3
	ldx	#<w2
	jsr	copy2w
	lda	#<reg3
	ldy	#>reg3
	ldx	#<w1
	jsr	copy2w
	jsr	sub	; subtract integer part from argument for fraction
	lda	w3+2
	bne	lab140
	sta	w3
lab140	rts


;
; isint - tests if w1 is an integer
; returns with Z=1 if yes, Z=0 if not
; uses w3
;
isint	jsr	int	; compute integer part...
	ldx	#7	; and compare it to argument
isnt01	lda	w1,x
	cmp	w3,x
	bne	isnt02
	dex
	bpl	isnt01
	lda	#0	; no difference found, w1 is integer
	rts
isnt02	lda	#$FF	; found a difference, not an integer
	rts


tanscale		; scale argument to range: 0 <= arg <= pi/4
	lda	#<twopi
	sta	ptr1
	lda	#>twopi
	sta	ptr1+1
	lda	#<scalereg	; put 2pi in scale register
	sta	ptr2
	lda	#>scalereg
	sta	ptr2+1
	jsr	copyreg
	jsr	scale	; get   2pi*frac(arg/2pi), -2pi <= arg <= 2pi
	lda	w3	; if arg < 0 add 2pi
	bpl	lab141
	lda	#<w3
	ldy	#>w3
	ldx	#<w1
	jsr	copy2w
	lda	#<twopi
	ldy	#>twopi
	ldx	#<w2
	jsr	copy2w
	jsr	add
lab141	lda	#<tpio2	; now find and save quadrant
	sta	ptr1
	lda	#>tpio2	;   ptr1        ptr2        n z
	sta	ptr1+1	;          =                x 1
	lda	#<w3	;          >                0 0
	sta	ptr2	;          <                1 0
	lda	#>w3
	sta	ptr2+1
	jsr	cmpreg	; 3pi/2 | w3
	bpl	lab142
	lda	#4	; w3 > 3pi/2
	sta	qdrnt
	jmp	tsc1
lab142	lda	#<pi
	sta	ptr1
	lda	#>pi
	sta	ptr1+1
	jsr	cmpreg	; pi | w3
	bpl	lab143
	lda	#3	; w3 > pi
	sta	qdrnt
	bra	tsc1
lab143	lda	#<pio2
	sta	ptr1
	lda	#>pio2
	sta	ptr1+1
	jsr	cmpreg	; pi/2 | w3
	bpl	lab144
	lda	#2	; w3 > pi/2
	sta	qdrnt
	bra	tsc1
lab144	lda	#1
	sta	qdrnt
tsc1	lda	qdrnt
	sta	ra
	cmp	#3
	bcc	lab145
	lda	#<w3	; map quadrants III and IV to I and II
	ldy	#>w3
	ldx	#<w1
	jsr	copy2w
	lda	#<pi
	ldy	#>pi
	ldx	#<w2
	jsr	copy2w
	jsr	sub
lab145	lda	qdrnt
	and	#1
	bne	lab146
	lda	#$80
	sta	tnsgn
lab146	lda	#<pio2
	sta	ptr1
	lda	#>pio2
	sta	ptr1+1
	jsr	cmpreg	; pi/2 | w3
	bpl	lab147
	lda	#<w3	; w3 > pi/2
	ldy	#>w3
	ldx	#<w1
	jsr	copy2w
	lda	#<pi
	ldy	#>pi
	ldx	#<w2
	jsr	copy2w
	jsr	sub
lab147	lda	w3	; range is: -pi/2 < arg < pi/2
	and	#$7f	; take abs value, tnsgn is already correct
	sta	w3
	lda	#<w3	; now range is:     0 <= arg <= pi/2.
	sta	ptr2	; reduce range to:  0 <= arg <= pi/4
	lda	#>w3
	sta	ptr2+1
	lda	#<pio4
	sta	ptr1
	lda	#>pio4
	sta	ptr1+1
	jsr	cmpreg	; pi/4 | w3
	bpl	lab148	; if upper part of range, pi/4 < arg < pi/2, is
	lda	#<pio2	;  mapped to lower part, 0 < arg < pi/4, toggle
	ldy	#>pio2	;  cotangent flag (cotflg).
	ldx	#<w1
	jsr	copy2w
	lda	#<w3
	ldy	#>w3
	ldx	#<w2
	jsr	copy2w
	jsr	sub
	lda	cotflg
	eor	#1
	sta	cotflg
lab148	rts		; all scaling is complete.

scale	lda	#<scalereg
	ldy	#>scalereg
	ldx	#<w2
	jsr	copy2w	; copy scale to w2
	jsr	div	; perform division
	lda	#<w3
	ldy	#>w3
	ldx	#<w1
	jsr	copy2w	; move result to w1
	jsr	frac	; take fractional portion of result
	lda	#<w3
	ldy	#>w3
	ldx	#<w1
	jsr	copy2w
	lda	#<scalereg
	ldy	#>scalereg
	ldx	#<w2
	jsr	copy2w	; copy scale to w2
	jmp	mul	; result is arg mod scale
			; jsr, rts


;
; square of w1 ->w3
;

pow2	lda	#<w1	; w1 -> w2
	ldy	#>w1
	ldx	#<w2
	jsr	copy2w
	jmp	mul	; and multiply
			; jsr, rts


;
; factorial of integer (0 <= w1 < 450) -> w3
;
fact	lda	w1+2	; filter out case where argument is zero
	beq	fct03	; if zero, result is 1.

	lda	w1
	and	#$80
	beq	fct00	; if number is negative, exit
	jmp	domnerr	; with domain error

fct00	lda	#<w1	; w1 -> w2
	ldy	#>w1
	ldx	#<w2
	jsr	copy2w

	jsr	isint	; check if argument is integer
	beq	fct05	; yes, OK
	jmp	domnerr	; error : argument is not integer

fct05	lda	#<w2	; less than 450 ?
	sta	ptr1
	lda	#>w2
	sta	ptr1+1
	lda	#<int450
	sta	ptr2
	lda	#>int450
	sta	ptr2+1
	jsr	cmpreg
	bmi	fct04	; yes, go on
	jmp	rangerr	; range error

fct04	lda	#<unit	; 1 -> w1
	ldy	#>unit
	ldx	#<w1
	jsr	copy2w

	lda	#<w1	; argument equal to 1 ?
	sta	ptr1
	lda	#>w1
	sta	ptr1+1
	lda	#<w2
	sta	ptr2
	lda	#>w2
	sta	ptr2+1
	jsr	cmpreg
	bne	fct01	; no : go to computation

fct03	lda	#<unit	; yes : 1 -> w3 and exit
	ldy	#>unit
	ldx	#<w3
	jmp	copy2w	; jsr, rts

fct01	jsr	mul	; w1 x w2 -> w3
	lda	#<w2	; w2 -> w1
	ldy	#>w2
	ldx	#<w1
	jsr	copy2w
	lda	#<w3	; w3 -> w4
	ldy	#>w3
	ldx	#<w4
	jsr	copy2w
	lda	#<unit	; 1 -> w2
	ldy	#>unit
	ldx	#<w2
	jsr	copy2w
	jsr	sub	; w1 - w2 -> w3

	lda	w2	; w2 = w3 (1) ? if yes, finished
	eor	#$80	; restore sign of unit constant
	sta	w2	; resulting from previous substraction
	lda	#<w2
	sta	ptr1
	lda	#>w2
	sta	ptr1+1
	lda	#<w3
	sta	ptr2
	lda	#>w3
	sta	ptr2+1
	jsr	cmpreg
	beq	fct02

	lda	#<w3	; w3 -> w2
	ldy	#>w3
	ldx	#<w2
	jsr	copy2w
	lda	#<w4	; w4 -> w1 and next muliply
	ldy	#>w4
	ldx	#<w1
	jsr	copy2w
	bra	fct01

fct02	lda	#<w4	; w4 -> w3
	ldy	#>w4
	ldx	#<w3
	jmp	copy2w	; jsr, rts


;
; combinations : w1 C w2 -> w3
; TO DO : check for over/underflow in loop

ncr	jsr	nxrarg	; verify that w1 & w2 are integers and w1 >= w2

	jsr	sub	; w1 - w2 -> w3 (X - Y -> w3)
	lda	w2	; restore w2 (Y) sign
	eor	#$80
	sta	w2
	lda	#<w3	; we need to choose the smallest value
	sta	ptr1
	lda	#>w3	; between Y and (X-Y)
	sta	ptr1+1
	lda	#<w2	; compare w3 and w2
	sta	ptr2
	lda	#>w2
	sta	ptr2+1
	jsr	cmpreg
	bmi	ncr02
	lda	#<w2	; w2 -> w4 Y is smaller
	ldy	#>w2
	bra	ncr03
ncr02	lda	#<w3	; w3 -> w4 (X-Y) is smaller
	ldy	#>w3
ncr03	ldx	#<w4
	jsr	copy2w

	lda	#<unit	; test whether w4 > 1, = 1, or = 0
	sta	ptr1
	lda	#>unit
	sta	ptr1+1
	lda	#<w4
	sta	ptr2
	lda	#>w4
	sta	ptr2+1
	jsr	cmpreg
	bmi	ncr05	; w4 > 1 : proceed to normal computation

	beq	ncr04	; w4 = 1 : w1 -> w3 and exit

	lda	w4+2	; test if w4 = 0
	bne	ncr05	; no : normal computation

	lda	#<unit	; yes : exit with w3 = 1
	ldy	#>unit
	ldx	#<w3
	jmp	copy2w	; jsr, rts

ncr04	lda	#<w1	; both args equal : exit with w3 = X
	ldy	#>w1
	ldx	#<w3
	jmp	copy2w	; jsr, rts
	

ncr05	lda	#<unit	; 1 -> w2
	ldy	#>unit
	ldx	#<w2
	jsr	copy2w
	jsr	sub
	lda	#<w4	; w4 -> w2
	ldy	#>w4
	ldx	#<w2
	jsr	copy2w
	lda	#<w3	; w3 -> w4
	ldy	#>w3
	ldx	#<w4
	jsr	copy2w
			; main loop
ncr00	jsr	div	; w1 / w2 -> w3
	lda	#<w3	; w3 -> w1
	ldy	#>w3
	ldx	#<w1
	jsr	copy2w
	lda	#<w4	; swap w2 and w4
	ldy	#>w4
	ldx	#<w2
	jsr	swap2w
	jsr	mul	; w1 * w2 -> w3
	lda	#<w3	; w3 -> w5
	ldy	#>w3
	ldx	#<w5
	jsr	copy2w
	lda	#<w2	; w2 -> w1
	ldy	#>w2
	ldx	#<w1
	jsr	copy2w
	lda	#<unit	; 1 -> w2
	ldy	#>unit
	ldx	#<w2
	jsr	copy2w
	jsr	sub	; w1 - w2 -> w3
	lda	#<w4	; w4 -> w1
	ldy	#>w4
	ldx	#<w1
	jsr	copy2w
	lda	#<w3	; w3 -> w4
	ldy	#>w3
	ldx	#<w4
	jsr	copy2w
	jsr	add	; w1 + w2 (-1) -> w3
	lda	#<w3	; w3 -> w1
	ldy	#>w3
	ldx	#<w1
	jsr	copy2w
	lda	w2	; restore sign of unit constant
	eor	#$80	; resulting from previous substraction
	sta	w2
	lda	#<w1	; compare w1 and w2
	sta	ptr1
	lda	#>w1
	sta	ptr1+1
	lda	#<w2
	sta	ptr2
	lda	#>w2
	sta	ptr2+1
	jsr	cmpreg
	beq	ncr01	; if eqal, we are done
	
	lda	#<w5	; w5 -> w1
	ldy	#>w5
	ldx	#<w1
	jsr	copy2w
	lda	#<w3	; w3 -> w2
	ldy	#>w3
	ldx	#<w2
	jsr	copy2w
	jmp	ncr00	; and next iteration

ncr01	lda	#<w5	; w5 -> w3
	ldy	#>w5
	ldx	#<w3
	jmp	copy2w	; jsr, rts


;
; permutations : w1 P w2 -> w3
; TO DO : check for over/underflow in loop

npr	jsr	nxrarg	; verify that w1 & w2 are integers and w1 >= w2

	lda	#<unit	; test for w2 > 1
	sta	ptr1
	lda	#>unit
	sta	ptr1+1
	lda	#<w2
	sta	ptr2
	lda	#>w2
	sta	ptr2+1
	jsr	cmpreg
	bmi	npr02	; w2 > 1 : proceed to normal computation

	lda	w2+2	; is w2 null ?
	bne	npr03	; no (then w2 = 1) : w1 -> w3 and exit

	lda	#<unit	; yes : exit with w3 = 1
	ldy	#>unit
	ldx	#<w3
	jmp	copy2w	; jsr, rts

npr03	lda	#<w1	; exit with w3 = X
	ldy	#>w1
	ldx	#<w3
	jmp	copy2w	; jsr, rts

npr02	lda	#<w1	; w1 -> w4
	ldy	#>w1
	ldx	#<w4
	jsr	copy2w
	lda	#<w1	; w1 -> w5
	ldy	#>w1
	ldx	#<w5
	jsr	copy2w
	jsr	sub	; X - Y
	lda	#<w3	; w3 -> w1
	ldy	#>w3
	ldx	#<w1
	jsr	copy2w
	lda	#<unit	; 1 -> w2
	ldy	#>unit
	ldx	#<w2
	jsr	copy2w
	jsr	add
	lda	#<w5	; w5 -> w1
	ldy	#>w5
	ldx	#<w1
	jsr	copy2w
	lda	#<w3	; w3 -> w5
	ldy	#>w3
	ldx	#<w5
	jsr	copy2w

npr00	lda	#<unit	; 1 -> w2
	ldy	#>unit
	ldx	#<w2
	jsr	copy2w
	jsr	sub	; decrement multiplicand
	lda	#<w4	; w4 -> w1
	ldy	#>w4
	ldx	#<w1
	jsr	copy2w
	lda	#<w3	; w3 -> w2
	ldy	#>w3
	ldx	#<w2
	jsr	copy2w
	jsr	mul
	lda	#<w3	; w3 -> w4
	ldy	#>w3
	ldx	#<w4
	jsr	copy2w
	lda	#<w2	; compare w2 and w5
	sta	ptr1
	lda	#>w2
	sta	ptr1+1
	lda	#<w5
	sta	ptr2
	lda	#>w5
	sta	ptr2+1
	jsr	cmpreg
	beq	npr01	; if equal, we're done

	lda	#<w2	; else w2 -> w1
	ldy	#>w2
	ldx	#<w1
	jsr	copy2w
	bra	npr00	; and loop

npr01	rts


;
; nxrarg - nCr and nPr arguments test
; w1 and w2 must be positive integers, else exit with domain error
; w1 must be greater than (or equal to) w2, else exit with range error
;

nxrarg	lda	w1	; verify w1 and w2 both positive
	ora	w2
	and	#$80
	beq	nxr01
	pla		; if not, drop return address to nXr
	pla
	jmp	domnerr	; and exit with domain error raised

nxr01	jsr	isint	; test if w1 is an integer
	beq	nxr02
	pla		; drop return address to nXr
	pla
	jmp	domnerr	; w1 not an integer : domain error

nxr02	lda	#<w1	; save w1 to w4
	ldy	#>w1
	ldx	#<w4
	jsr	copy2w
	lda	#<w2	; copy w2 to w1
	ldy	#>w2
	ldx	#<w1
	jsr	copy2w
	jsr	isint	; and test if integer
	beq	nxr03
	pla		; drop return address to nXr
	pla
	jmp	domnerr	; w2 not an integer : domain error

nxr03	lda	#<w4	; get back w1 from w4
	ldy	#>w4
	ldx	#<w1
	jsr	copy2w
	lda	#<w1	; now we test if w1 > w2
	sta	ptr1
	lda	#>w1
	sta	ptr1+1
	lda	#<w2
	sta	ptr2
	lda	#>w2
	sta	ptr2+1
	jsr	cmpreg

	bpl	nxr04	; if w1 < w2 ...

	pla		; ... drop return address to nXr
	pla
	jmp	rangerr	; and exit with range error raised

nxr04	rts		; else : w1 > or = to w2 : go to computation


;
; cubic root : w1 ^ (1/3) -> w3
;

cubrt	lda	w1+2	; w1 null ?
	bne	cbrt00
	jmp	retzr	; yes : exit with zero

cbrt00	lda	w1	; we need to save w1 mantissa sign
	and	#$80
	pha
	lda	w1
	and	#$7F	; we work on absolute value
	sta	w1
	lda	#<rcp3	; get 1/3 in w2
	ldy	#>rcp3
	ldx	#<w2
	jsr	copy2w
	jsr	pow	; and compute power
	pla		; restore sign of w1 in w3
	beq	cbrt01
	ora	w3
	sta	w3
cbrt01	rts


;
; p10ow : 10 ^ w1 -> w3
;

p10ow	lda	#<w1	; copy w1 to w2
	ldy	#>w1
	ldx	#<w2
	jsr	copy2w
	lda	#<unit	; copy unit constant to w1
	ldy	#>unit
	ldx	#<w1
	jsr	copy2w
	inc	w1+1	; set exponent to 1 : 10 constant
	jmp	pow	; and compute power
			; jsr, rts


;
; p2r : polar to rectangular coordinates conversion
; Specific points (shared with next function r2p) :
; 1) On entry, only w1 is loaded with X stack register
; 2) This function and r2p are the only ones
; producing results both in X and Y.
;

p2r	lda	w1+2		; radius null ?
	bne	p2r_00

	jsr	retzr		; yes : 0 -> w3

	lda	#<w3		; 0 -> Y register
	ldy	#>w3
	ldx	#<st_Y
	jmp	copy2w		; jsr, rts

p2r_00	lda	#<w1		; non-zero radius -> w5
	ldy	#>w1
	ldx	#<w5
	jsr	copy2w

	lda	#<st_Y		; theta -> w1
	ldy	#>st_Y
	ldx	#<w1
	jsr	copy2w

	jsr	sin		; compute sine

	lda	#<w3		; and copy in w1
	ldy	#>w3
	ldx	#<w1
	jsr	copy2w

	lda	#<w5		; get back radius in w2
	ldy	#>w5
	ldx	#<w2
	jsr	copy2w

	jsr	mul		; multiply : y is in w3

	lda	#<st_Y		; theta -> w1
	ldy	#>st_Y
	ldx	#<w1
	jsr	copy2w

	lda	#<w3		; y -> Y register
	ldy	#>w3
	ldx	#<st_Y
	jsr	copy2w

	jsr	cos		; compute cosine

	lda	#<w3		; and copy in w1
	ldy	#>w3
	ldx	#<w1
	jsr	copy2w

	lda	#<w5		; get back radius
	ldy	#>w5
	ldx	#<w2
	jsr	copy2w

	jmp	mul		; x is in w3
				; jsr, rts


;
; r2p : rectangular to polar coordinates conversion
; Beware of some peculiarities - see p2r
;

r2p	lda	w1+2		; we first test x / y = 0 / 0
	ora	st_Y+2
	bne	r2p_00

	jmp	retzr		; both are null, update w3 <- 0
				; jsr, rts

r2p_00	lda	w1+2		; next we compute the radius
	bne	r2p_01		; if X is not null, square it

	lda	#<w1		; else simply copy it to w5
	ldy	#>w1
	bra	r2p_02

r2p_01	jsr	pow2		; X ^ 2

	lda	#<w3		; and copy in w5
	ldy	#>w3
r2p_02	ldx	#<w5
	jsr	copy2w

	lda	st_Y+2		; if Y null, no need to square it
	beq	r2p_03		; nor to add

	lda	#<st_Y		; Y -> w1
	ldy	#>st_Y
	ldx	#<w1
	jsr	copy2w

	jsr	pow2		; Y ^ 2

	lda	#<w5		; w5 (X ^ 2) -> w1
	ldy	#>w5
	ldx	#<w1
	jsr	copy2w

	lda	#<w3		; w3 (Y ^ 2) -> w2
	ldy	#>w3
	ldx	#<w2
	jsr	copy2w

	jsr	add		; X ^ 2 + Y ^ 2 -> w3

r2p_03	lda	#<w3		; w3 -> w1
	ldy	#>w3
	ldx	#<w1
	jsr	copy2w

	jsr	sqrt		; this ends the radius computation

	lda	#<w3		; radius -> w5
	ldy	#>w3
	ldx	#<w5
	jsr	copy2w

	lda	st_X+2		; is X null ?
	bne	r2p_04
				; yes : we already know that theta
	lda	#<pio2		; has an absolute value of pi / 2
	ldy	#>pio2
	ldx	#<w3
	jsr	copy2w

	lda	st_Y		; next we copy Y sign into w3's
	and	#$80
	ora	w3
	sta	w3

	bra	r2p_98		; on to final transfer in X / Y

r2p_04	lda	#<st_Y		; X not null : compute arctg(Y/X)
	ldy	#>st_Y
	ldx	#<w1
	jsr	copy2w

	lda	#<st_X
	ldy	#>st_X
	ldx	#<w2
	jsr	copy2w

	jsr	div		; Y / X -> w3

	lda	#<w3
	ldy	#>w3
	ldx	#<w1
	jsr	copy2w

	jsr	atan		; arctg(Y/X) -> w3 (theta)

	lda	st_X		; is X positive ?
	and	#$80
	beq	r2p_98		; yes : do the final transfer

	lda	#<w3		; no : we need to add or substract pi
	ldy	#>w3		; according to Y sign
	ldx	#<w1
	jsr	copy2w		; theta -> w1

	lda	#<pi
	ldy	#>pi
	ldx	#<w2
	jsr	copy2w		; pi ->  w2

	lda	st_Y		; copy Y sign to w2's
	and	#$80
	ora	w2
	sta	w2

	jsr	add		; pi is added or substracted

r2p_98	lda	#<w3		; theta -> Y register
	ldy	#>w3
	ldx	#<st_Y
	jsr	copy2w

	lda	#<w5		; radius -> w3
	ldy	#>w5
	ldx	#<w3
	jmp	copy2w
				; jsr, rts


;
; hr2hms : conversion from decimal hours to hours/minutes/seconds
;
; We use the formula hms(x) = (90.x + 100.int(x) + int(60.x)) / 250
;

hr2hms	lda	w1+2	; w1 null ?
	bne	hms00
	jmp	retzr	; yes : exit with zero

	jsr	isint	; w1 not null : integer ?
	bne	hms00
	rts		; yes, nothing to do

hms00	lda	w1	; we need to save w1 mantissa sign
	and	#$80
	pha
	lda	w1
	and	#$7F	; we work on absolute value
	sta	w1

	lda	#<w1	; w1 (hr) -> w5
	ldy	#>w1
	ldx	#<w5
	jsr	copy2w

	lda	#<int90	; 90 -> w2
	ldy	#>int90
	ldx	#<w2
	jsr	copy2w

	jsr	mul	; w3 <- 90.w1

	lda	#<w3	; w3 -> w4
	ldy	#>w3
	ldx	#<w4
	jsr	copy2w

	jsr	int	; w3 <- int(x)

	lda	#<w3	; w3 -> w1
	ldy	#>w3
	ldx	#<w1
	jsr	copy2w

	lda	#<int100	; 100 -> w2
	ldy	#>int100
	ldx	#<w2
	jsr	copy2w

	jsr	mul	; w3 <- 100.int(x)

	lda	#<w3	; w3 -> w2
	ldy	#>w3
	ldx	#<w2
	jsr	copy2w

	lda	#<w4	; w4 -> w1
	ldy	#>w4
	ldx	#<w1
	jsr	copy2w

	jsr	add	; w3 <- 90.x + 100.int(x)

	lda	#<w3	; w3 -> w4
	ldy	#>w3
	ldx	#<w4
	jsr	copy2w

	lda	#<w5	; w5 -> w1
	ldy	#>w5
	ldx	#<w1
	jsr	copy2w

	lda	#<int60	; 60 -> w2
	ldy	#>int60
	ldx	#<w2
	jsr	copy2w

	jsr	mul	; w3 <- 60.x

	lda	#<w3	; w3 -> w1
	ldy	#>w3
	ldx	#<w1
	jsr	copy2w

	jsr	int	; w3 <- int(60.x)

	lda	#<w4	; w4 -> w1
	ldy	#>w4
	ldx	#<w1
	jsr	copy2w

	lda	#<w3	; w3 -> w2
	ldy	#>w3
	ldx	#<w2
	jsr	copy2w

	jsr	add	; w3 <- 90.x + 100.int(x) + int(60.x)

	lda	#<w3	; w3 -> w1
	ldy	#>w3
	ldx	#<w1
	jsr	copy2w

	lda	#<int250	; 250 -> w2
	ldy	#>int250
	ldx	#<w2
	jsr	copy2w

	jsr	div	; w3 <- final result

	pla		; restore sign of w1 in w3
	beq	hms01

	ora	w3
	sta	w3

hms01	rts


;
; hms2hr : conversion from hours/minutes/seconds to decimal hours
;
; We use the formula hr(x) = (250.x - 60.int(x) - int(100.x)) / 90
;

hms2hr	lda	w1+2	; w1 null ?
	bne	hr00
	jmp	retzr	; yes : exit with zero

	jsr	isint	; w1 not null : integer ?
	bne	hr00
	rts		; yes, nothing to do

hr00	lda	w1	; we need to save w1 mantissa sign
	and	#$80
	pha
	lda	w1
	and	#$7F	; we work on absolute value
	sta	w1

	lda	#<w1	; w1 (hms) -> w5
	ldy	#>w1
	ldx	#<w5
	jsr	copy2w

	lda	#<int250	; 250 -> w2
	ldy	#>int250
	ldx	#<w2
	jsr	copy2w

	jsr	mul	; w3 <- 250.x

	lda	#<w3	; w3 -> w4
	ldy	#>w3
	ldx	#<w4
	jsr	copy2w

	jsr	int	; w3 <- int(x)

	lda	#<w3	; w3 -> w1
	ldy	#>w3
	ldx	#<w1
	jsr	copy2w

	lda	#<int60	; 60 -> w2
	ldy	#>int60
	ldx	#<w2
	jsr	copy2w

	jsr	mul	; w3 <- 60.int(x)

	lda	#<w4	; w4 -> w1
	ldy	#>w4
	ldx	#<w1
	jsr	copy2w

	lda	#<w3	; w3 -> w2
	ldy	#>w3
	ldx	#<w2
	jsr	copy2w

	jsr	sub	; w3 <- 250.x - 60.int(x)

	lda	#<w3	; w3 -> w4
	ldy	#>w3
	ldx	#<w4
	jsr	copy2w

	lda	#<w5	; w5 -> w1
	ldy	#>w5
	ldx	#<w1
	jsr	copy2w

	lda	#<int100	; 100 -> w2
	ldy	#>int100
	ldx	#<w2
	jsr	copy2w

	jsr	mul	; w3 <- 100.x

	lda	#<w3	; w3 -> w1
	ldy	#>w3
	ldx	#<w1
	jsr	copy2w

	jsr	int	; w3 <- int(100.x)

	lda	#<w4	; w4 -> w1
	ldy	#>w4
	ldx	#<w1
	jsr	copy2w

	lda	#<w3	; w3 -> w2
	ldy	#>w3
	ldx	#<w2
	jsr	copy2w

	jsr	sub	; w3 <- 250.x - 60.int(x) - int(100.x)

	lda	#<w3	; w3 -> w1
	ldy	#>w3
	ldx	#<w1
	jsr	copy2w

	lda	#<int90	; 90 -> w2
	ldy	#>int90
	ldx	#<w2
	jsr	copy2w

	jsr	div	; w3 <- final result

	pla		; restore sign of w1 in w3
	beq	hr01

	ora	w3
	sta	w3

hr01	rts


;
; compare registers pointed to by ptr1 and ptr2.
; returns the following:
;                                       n z
;   zflag = 1 if numbers are the same   x 1
;   nflag = 0 if ptr1 > ptr2            0 0
;   nflag = 1 if ptr1 < ptr2            1 0
;
cmpreg	sed
	ldy	#0
	lda	(ptr1),y
	eor	(ptr2),y
	bpl	lab149
;
;   mantissa signs differ, positive number is larger
;
	lda	(ptr1),y	; set nflag using ptr1
	ora	#1	; make sure zflag = 0
	rts
;
;   mantissa signs are the same, check exponent signs
;
lab149	asl		; shift exponent sign into sign bit
	bpl	lab150
;
;   exponent signs differ, larger number has:
;       mantissa sign positive, exponent sign positive
;       mantissa sign negative, exponent sign negative
;
	lda	(ptr1),y
	bmi	negm
	ldy	#2
	lda	(ptr1),y
	bne	nz1
	lda	#$80	; ptr1 = 0 and ptr1 = frac
	rts
nz1	lda	(ptr2),y
	bne	nz2
	rts		; ptr2 = 0 and ptr1 = frac
nz2	ldy	#0
	lda	(ptr1),y
	asl
	ora	#$1
	rts
negm	lda	(ptr2),y
	asl
	ora	#$1
	rts
;
;   mantissa exponent signs are the same, check exponent magnitude
;
lab150	lda	(ptr1),y
	eor	(ptr2),y
	bne	lab151
;
;   first exponent digit matches, check next two
;
	iny
	lda	(ptr1),y
	eor	(ptr2),y
	bne	lab151
;
;   all signs and exponents match, compare mantissas
;
	ldy	#7
	ldx	#5
	stz	tmp1
	sec
mantst	lda	(ptr2),y
	sbc	(ptr1),y
	beq	zr
	inc	tmp1
zr	dey
	dex
	bpl	mantst

	php			; save test result, check mantissa sign
	ldy	#0		; (if negative, absolute values compare
	lda	(ptr1),y	; the reverse way)
	and	#$80
	beq	mants1

	lda	tmp1		; negative
	bne	mants2

	plp			; equal
	lda	#0
	rts

mants2	plp
	bcs	p1max
	bra	p2max
	rts

mants1	plp			; positive
	bcc	p1max
	lda	tmp1
	bne	p2max
	rts
;
;   all signs match, but exponents differ
;
;   largest number
;       msgn+, esgn+, exponent largest   80.0 >  1.23 
;       msgn+, esgn-, exponent smallest  0.33 >  0.02
;       msgn-, esgn+, exponent smallest -1.23 > -80.0
;       msgn-, esgn-, exponent largest  -0.02 > -0.33
;
lab151	ldy	#1
	sec
	lda	(ptr2),y
	sbc	(ptr1),y
	dey
	lda	(ptr2),y
	sbc	(ptr1),y
	bcs	e2max
;   ptr1 has largest exponent, all signs are the same
	lda	(ptr1),y
	and	#$c0
	beq	p1max
	cmp	#$c0
	beq	p1max
p2max	lda	#$81
	rts
p1max	lda	#1
	rts
;   ptr2 has largest exponent, all signs are the same
e2max	lda	(ptr2),y
	and	#$c0
	beq	p2max
	cmp	#$c0
	beq	p2max
	bne	p1max


;
;   return with zero
;
;   utility routine to place zero in w3 and return
;
;retzr	lda	#<zero		; 133 cycles
;	ldy	#>zero
;	ldx	#<w3
;	jmp	copy2w	; jsr, rts

retzr	lda	#0		; 89 cycles
	ldx	#7
rtz00	sta	w3,x		; zero value is 8 $00 bytes
	dex
	bpl	rtz00
	rts


;
;   domain error handler
;
;   returns domain error string with error flags set
;
;       $30,2,0,0,0,0,0,0
;
domnerr	lda	#<dmnerr
	ldy	#>dmnerr
	bra	erhndlr


;
;   range error handler
;
;   returns range error string with error flags set
;
;       $30,1,0,0,0,0,0,0
;
rangerr	lda	#<rngerr
	ldy	#>rngerr
	bra	erhndlr


;
;   overflow error handler
;
;   returns overflow error string with error flags set
;
;       $29,$99,$99,$99,$99,$99,$99,$99
;
ovrferr	lda	#<ovferr
	ldy	#>ovferr
	bra	erhndlr


;
;   divide-by-zero error handler
;
;   returns divide-by-zero error string with error flags set
;
;       $30,3,0,0,0,0,0,0
;
dvzrerr	lda	#<dvzerr
	ldy	#>dvzerr
;	bra	erhndlr

; !! DO NOT INSERT INYTHING HERE !!

;
;   general error handler
;
;   returns error string in w3 and executes 'brk' instruction
;

erhndlr	ldx	#<w3		; put error code in w3
	jmp	copy2w		; jsr, rts



; -------------------------------------------------------------
;
;   utility routines
;


; copy number from (ptr1) to (ptr2)
; enter with src register page in 'y', src register offset in 'a',
; dest register offset in 'x' (assumed destination page is for 'wx' regs).
;
copy2w	sty	ptr1+1
	sta	ptr1
	stx	ptr2
	lda	#>w1	; get page used for working registers
	sta	ptr2+1
	ldy	#7
lab004	lda	(ptr1),y
	sta	(ptr2),y
	dey
;	bpl	lab004
;	rts

; unrolling the loop : 124 cycles instead of 149
	lda	(ptr1),y
	sta	(ptr2),y
	dey
	lda	(ptr1),y
	sta	(ptr2),y
	dey
	lda	(ptr1),y
	sta	(ptr2),y
	dey
	lda	(ptr1),y
	sta	(ptr2),y
	dey
	lda	(ptr1),y
	sta	(ptr2),y
	dey
	lda	(ptr1),y
	sta	(ptr2),y
	dey
	lda	(ptr1),y
	sta	(ptr2),y
	rts


; swap numbers (ptr1) and (ptr2)
; enter with first register page in 'y', first register offset in 'a',
; second register offset in 'x'
; we assume that (ptr2) is in the 'wx' registers page
;
swap2w	sty	ptr1+1
	sta	ptr1
	stx	ptr2
	lda	#>w1
	sta	ptr2+1
	ldy	#7		; loop unrolled : 293 -> 244 cycles
swp00	lda	(ptr1),y	; (and pha/pla replaced with tax/txa)
	tax
	lda	(ptr2),y
	sta	(ptr1),y
	txa
	sta	(ptr2),y
	dey
	lda	(ptr1),y
	tax
	lda	(ptr2),y
	sta	(ptr1),y
	txa
	sta	(ptr2),y
	dey
	lda	(ptr1),y
	tax
	lda	(ptr2),y
	sta	(ptr1),y
	txa
	sta	(ptr2),y
	dey
	lda	(ptr1),y
	tax
	lda	(ptr2),y
	sta	(ptr1),y
	txa
	sta	(ptr2),y
	dey
	lda	(ptr1),y
	tax
	lda	(ptr2),y
	sta	(ptr1),y
	txa
	sta	(ptr2),y
	dey
	lda	(ptr1),y
	tax
	lda	(ptr2),y
	sta	(ptr1),y
	txa
	sta	(ptr2),y
	dey
	lda	(ptr1),y
	tax
	lda	(ptr2),y
	sta	(ptr1),y
	txa
	sta	(ptr2),y
	dey
	lda	(ptr1),y
	tax
	lda	(ptr2),y
	sta	(ptr1),y
	txa
	sta	(ptr2),y
	rts


;
; Register copy - usable anywhere in memory
; !! BEWARE OF PAGE BOUNDARIES !!
;
copyreg	ldy	#7
lab152	lda	(ptr1),y
	sta	(ptr2),y
	dey
	bpl	lab152
	rts


;
; rr_clr - clears working registers ra to rf
; (6 x 16 bytes)
;
rr_clr	ldx	#$5F
	lda	#0
rclr00	sta	ra,x
	dex
	bpl	rclr00
	rts



;-------------------------------------------------------------------------
; Constants
;
; functions addresses jump table
; BCD operations constants table
; floating-point constants

;
;   The following two tables are used in shifting a BCD register
;   one BCD digit to the right or left. 
;
;
;   'tlh' maps the lower decimal digit of its BCD index into
;   the high digit.
;
;   Example:
;
;           ldy #$35
;           lda tlh,y
;
;   Now the 'a' register contains #$50; i.e. the '5' in '35'
;   is shifted into the high digit position and the low
;   digit position is cleared.
;
;   This table, and the one following, support high speed
;   shifting of multibyte BCD registers.
;
tlh .byte 0,$10,$20,$30,$40,$50,$60,$70,$80,$90,0,0,0,0,0,0
    .byte 0,$10,$20,$30,$40,$50,$60,$70,$80,$90,0,0,0,0,0,0
    .byte 0,$10,$20,$30,$40,$50,$60,$70,$80,$90,0,0,0,0,0,0
    .byte 0,$10,$20,$30,$40,$50,$60,$70,$80,$90,0,0,0,0,0,0
    .byte 0,$10,$20,$30,$40,$50,$60,$70,$80,$90,0,0,0,0,0,0
    .byte 0,$10,$20,$30,$40,$50,$60,$70,$80,$90,0,0,0,0,0,0
    .byte 0,$10,$20,$30,$40,$50,$60,$70,$80,$90,0,0,0,0,0,0
    .byte 0,$10,$20,$30,$40,$50,$60,$70,$80,$90,0,0,0,0,0,0
    .byte 0,$10,$20,$30,$40,$50,$60,$70,$80,$90,0,0,0,0,0,0
    .byte 0,$10,$20,$30,$40,$50,$60,$70,$80,$90,0,0,0,0,0,0
;
;   'thl' maps the upper digit of its BCD index into the
;   low digit.
;
;   Example:
;
;           ldy #$35
;           lda thl,y
;
;   Now the 'a' register contains #$03; i.e. the '3' in '35'
;   is shifted to the low digit position and the high digit
;   position is cleared.
;
thl .byte 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0
    .byte 1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1
    .byte 2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2
    .byte 3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3
    .byte 4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4
    .byte 5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5
    .byte 6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6
    .byte 7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7
    .byte 8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8
    .byte 9,9,9,9,9,9,9,9,9,9,9,9,9,9,9,9
;
;   'dec2bin' maps a BCD value into its equivalent binary value.
;
;   Example:
;
;           ldy #$35            ; '35' decimal
;           lda dec2bin,y
;
;   The 'a' register now contains 23h = 0010 0011b = 35 decimal
;
dec2bin
    .byte 0,1,2,3,4,5,6,7,8,9,0,0,0,0,0,0
    .byte $0a,$0b,$0c,$0d,$0e,$0f,$10,$11,$12,$13,0,0,0,0,0,0
    .byte $14,$15,$16,$17,$18,$19,$1a,$1b,$1c,$1d,0,0,0,0,0,0
    .byte $1e,$1f,$20,$21,$22,$23,$24,$25,$26,$27,0,0,0,0,0,0
    .byte $28,$29,$2a,$2b,$2c,$2d,$2e,$2f,$30,$31,0,0,0,0,0,0
    .byte $32,$33,$34,$35,$36,$37,$38,$39,$3a,$3b,0,0,0,0,0,0
    .byte $3c,$3d,$3e,$3f,$40,$41,$42,$43,$44,$45,0,0,0,0,0,0
    .byte $46,$47,$48,$49,$4a,$4b,$4c,$4d,$4e,$4f,0,0,0,0,0,0
    .byte $50,$51,$52,$53,$54,$55,$56,$57,$58,$59,0,0,0,0,0,0
    .byte $5a,$5b,$5c,$5d,$5e,$5f,$60,$61,$62,$63,0,0,0,0,0,0
;
;   'bin2dec' maps a binary value into its BCD equivalent value.
;
;   Example:
;
;           ldy #$2c            ; '2c' hex = 0010 1100b = 44 decimal
;           lda bin2dec,y
;
;   The 'a' register now contains #$44.
;
bin2dec
    .byte 0,1,2,3,4,5,6,7,8,9
    .byte $10,$11,$12,$13,$14,$15,$16,$17,$18,$19
    .byte $20,$21,$22,$23,$24,$25,$26,$27,$28,$29
    .byte $30,$31,$32,$33,$34,$35,$36,$37,$38,$39
    .byte $40,$41,$42,$43,$44,$45,$46,$47,$48,$49
    .byte $50,$51,$52,$53,$54,$55,$56,$57,$58,$59
    .byte $60,$61,$62,$63,$64,$65,$66,$67,$68,$69
    .byte $70,$71,$72,$73,$74,$75,$76,$77,$78,$79
    .byte $80,$81,$82,$83,$84,$85,$86,$87,$88,$89
    .byte $90,$91,$92,$93,$94,$95,$96,$97,$98,$99 
;
;   srtbl supports the division of BCD numbers by two
;
;   To use, shift the BCD register right one bit and
;   then replace each byte with the value take from
;   this lookup table using the existing byte as an
;   index.
;
srtbl   ; shift right table
    .byte  $00,$01,$02,$03,$04,$05,$06,$07,$05,$06,$07,$08,$09,0,0,0
    .byte  $10,$11,$12,$13,$14,$15,$16,$17,$15,$16,$17,$18,$19,0,0,0
    .byte  $20,$21,$22,$23,$24,$25,$26,$27,$25,$26,$27,$28,$29,0,0,0
    .byte  $30,$31,$32,$33,$34,$35,$36,$37,$35,$36,$37,$38,$39,0,0,0
    .byte  $40,$41,$42,$43,$44,$45,$46,$47,$45,$46,$47,$49,$49,0,0,0
    .byte  $50,$51,$52,$53,$54,$55,$56,$57,$55,$56,$57,$58,$59,0,0,0
    .byte  $60,$61,$62,$63,$64,$65,$66,$67,$65,$66,$67,$68,$69,0,0,0
    .byte  $70,$71,$72,$73,$74,$75,$76,$77,$75,$76,$77,$78,$79,0,0,0
    .byte  $50,$51,$52,$53,$54,$55,$56,$57,$55,$56,$57,$58,$59,0,0,0
    .byte  $60,$61,$62,$63,$64,$65,$66,$67,$65,$66,$67,$68,$69,0,0,0
    .byte  $70,$71,$72,$73,$74,$75,$76,$77,$75,$76,$77,$78,$79,0,0,0
    .byte  $80,$81,$82,$83,$84,$85,$86,$87,$85,$86,$87,$88,$89,0,0,0
    .byte  $90,$91,$92,$93,$94,$95,$96,$97,$95,$96,$97,$98,$99,0,0,0

sltbl   ; shift left table
;    .byte $00,$01,$02,$03,$04,$08,$09,$0a,$0b,$0c,0,0,0,0,0,0
;    .byte  $10,$11,$12,$13,$14,$18,$19,$1a,$1b,$1c,0,0,0,0,0,0
;    .byte  $20,$21,$22,$23,$24,$28,$29,$2a,$2b,$2c,0,0,0,0,0,0
;    .byte  $30,$31,$32,$33,$34,$38,$39,$3a,$3b,$3c,0,0,0,0,0,0
;    .byte  $40,$41,$42,$43,$44,$48,$49,$4a,$4b,$4c,0,0,0,0,0,0
;    .byte  $80,$81,$82,$83,$84,$88,$89,$8a,$8b,$8c,0,0,0,0,0,0
;    .byte  $90,$91,$92,$93,$94,$98,$99,$9a,$9b,$9c,0,0,0,0,0,0
;    .byte  $a0,$a1,$a2,$a3,$a4,$a8,$a9,$aa,$ab,$ac,0,0,0,0,0,0
;    .byte  $b0,$b1,$b2,$b3,$b4,$b8,$b9,$ba,$bb,$bc,0,0,0,0,0,0
;    .byte  $c0,$c1,$c2,$c3,$c4,$c8,$c9,$ca,$cb,$cc,0,0,0,0,0,0

hex2asc
;    .byte '0','1','2','3','4','5','6','7','8','9','A','B','C','D','E','F'
;    .BYTE "0123456789ABCDEF"



;
;   miscellaneous constants in floating
;   point format
;
zero   .byte $00,$00,$00,$00,$00,$00,$00,$00
ln2    .byte $40,$01,$69,$31,$47,$18,$05,$60
ln10   .byte $00,$00,$23,$02,$58,$50,$92,$99
iln10  .byte $40,$01,$43,$42,$94,$48,$19,$03
;exp0   .byte $00,$00,$27,$18,$28,$18,$28,$46
;sqt2   .byte $00,$00,$14,$14,$21,$35,$62,$37
pi     .byte $00,$00,$31,$41,$59,$26,$53,$59
twopi  .byte $00,$00,$62,$83,$18,$53,$07,$18
;sqp5   .byte $40,$01,$70,$71,$06,$78,$11,$87
;invpi  .byte $40,$01,$31,$83,$09,$88,$61,$84
;i2pi   .byte $40,$01,$15,$91,$54,$94,$30,$92	; unused ?
pio2   .byte $00,$00,$15,$70,$79,$63,$26,$79
;npio2  .byte $80,$00,$15,$70,$79,$63,$26,$79	; unused ?
tpio2  .byte $00,$00,$47,$12,$38,$89,$80,$38
pio4   .byte $40,$01,$78,$53,$98,$16,$33,$98
unit   .byte $00,$00,$10,$00,$00,$00,$00,$00
half   .byte $40,$01,$50,$00,$00,$00,$00,$00
;nrpi2  .byte $00,$15,$70,$79,$00,$00,$00,$00	; unused ?
;hnth   .byte $40,$02,$10,$00,$00,$00,$00,$00	; unused ?
rcp3   .byte $40,$01,$33,$33,$33,$33,$33,$33
int60  .byte $00,$01,$60,$00,$00,$00,$00,$00
int90  .byte $00,$01,$90,$00,$00,$00,$00,$00
int100 .byte $00,$02,$10,$00,$00,$00,$00,$00
int250 .byte $00,$02,$25,$00,$00,$00,$00,$00
int450 .byte $00,$02,$45,$00,$00,$00,$00,$00
ovferr .byte $29,$99,$99,$99,$99,$99,$99,$99
rngerr .byte $30,1,0,0,0,0,0,0
dmnerr .byte $30,2,0,0,0,0,0,0
dvzerr .byte $30,3,0,0,0,0,0,0
;
;   atn(1), atn(0.1), atn(0.01), atn(0.001), ...
;
;   This table is in fixed point format
;
kxatn .byte $07,$85,$39,$81,$63,$39,$74,$48
      .byte $00,$99,$66,$86,$52,$49,$11,$62
      .byte $00,$09,$99,$96,$66,$68,$66,$65
      .byte $00,$00,$99,$99,$99,$66,$66,$67
      .byte $00,$00,$09,$99,$99,$99,$96,$67
      .byte $00,$00,$01,$00,$00,$00,$00,$00
      .byte $00,$00,$00,$10,$00,$00,$00,$00
      .byte $00,$00,$00,$01,$00,$00,$00,$00
      .byte $00,$00,$00,$00,$10,$00,$00,$00
      .byte $00,$00,$00,$00,$01,$00,$00,$00
      .byte $00,$00,$00,$00,$00,$10,$00,$00
      .byte $00,$00,$00,$00,$00,$01,$00,$00
;
;   ln(2), ln(1.1), ln(1.01),ln(1.001),...
;
;   This table is in fixed decimal format
;
kxlog .byte $06,$93,$14,$71,$80,$55,$99,$45
      .byte $00,$95,$31,$01,$79,$80,$43,$25
      .byte $00,$09,$95,$03,$30,$85,$31,$68
      .byte $00,$00,$99,$95,$00,$33,$30,$84
      .byte $00,$00,$09,$99,$95,$00,$03,$33
      .byte $00,$00,$00,$99,$99,$95,$00,$00
      .byte $00,$00,$00,$09,$99,$99,$95,$00
      .byte $00,$00,$00,$00,$99,$99,$99,$95
      .byte $00,$00,$00,$00,$10,$00,$00,$00
      .byte $00,$00,$00,$00,$01,$00,$00,$00
      .byte $00,$00,$00,$00,$00,$10,$00,$00
      .byte $00,$00,$00,$00,$00,$01,$00,$00
      .byte $00,$00,$00,$00,$00,$00,$10,$00
      .byte $00,$00,$00,$00,$00,$00,$01,$00
      .byte $00,$00,$00,$00,$00,$00,$00,$10
      .byte $00,$00,$00,$00,$00,$00,$00,$01
kln   .byte $C0,$01,$25,$00,$00,$00,$00,$00
      .byte $40,$01,$33,$33,$33,$33,$33,$33
      .byte $C0,$01,$50,$00,$00,$00,$00,$00
      .byte $00,$00,$10,$00,$00,$00,$00,$00
tst1  .byte $00,$01,$80,$00,$00,$00,$00,$00
tst2  .byte $00,$01,$80,$00,$00,$00,$00,$01
tst3  .byte $80,$01,$80,$00,$00,$00,$00,$00
tst4  .byte $80,$00,$12,$30,$00,$00,$00,$00
tst5  .byte $40,$01,$33,$00,$00,$00,$00,$00
tst6  .byte $40,$02,$20,$00,$00,$00,$00,$00
tst7  .byte $C0,$01,$33,$00,$00,$00,$00,$00
tst8  .byte $C0,$02,$20,$00,$00,$00,$00,$00

